"""Fig 2 figure
"""
import numpy as np
import matplotlib.pyplot as plt
#from scipy.stats import mstats
import pandas as pd
import geopandas as gpd
from scipy import stats
from shapely.geometry import Point
import matplotlib.pyplot as plt
from collections import defaultdict
from matplotlib.colors import Normalize

from energy_demand.plotting import result_mapping
from energy_demand.technologies import tech_related
from energy_demand.plotting import basic_plot_functions

def run(
        data_input,
        regions,
        simulation_yr_to_plot,
        population,
        fueltype_str,
        path_shapefile,
        fig_name
    ):
    """
    """
    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    # -----------------------------------------------------------
    # Iterate overall weather_yrs and store data in dataframe
    # (columns = timestep, rows: value of year)
    # -----------------------------------------------------------

    # List of selected data for every weather year and region (which is then converted to array)
    weather_yrs_data = defaultdict(dict)

    print("Weather yrs: " + str(list(data_input.keys())), flush=True)

    for weather_yr, data_weather_yr in data_input.items():

        # Weather year specific data for every region
        regions_fuel = data_weather_yr[simulation_yr_to_plot][fueltype_int]     # Select fueltype

        for region_nr, region_name in enumerate(regions):
            try:
                weather_yrs_data[region_name].append(regions_fuel[region_nr])
            except (KeyError, AttributeError):
                weather_yrs_data[region_name] = [regions_fuel[region_nr]]

    regional_statistics_columns = [
        'name',
        'mean_peak_h',
        'diff_av_max',
        'mean_peak_h_pp',
        'diff_av_max_pp',
        'std_dev_average_every_h',
        'std_dev_peak_h_norm_pop']

    df_stats = pd.DataFrame(columns=regional_statistics_columns)

    for region_name, region_data in weather_yrs_data.items():

        # Convert regional data to dataframe
        region_data_array = np.array(region_data)
        df = pd.DataFrame(
            region_data_array,
            columns=range(8760))

        # Calculate regional statistics
        mean = df.mean(axis=0)
        std_dev = df.std(axis=0) #standard deviation across every hour

        # Get maximum per colum
        #max_every_h = df.max()
        #colum_max_h = max_every_h.argmax() #get colum (respesctively hour) of maximum value

        # Average standard deviation across every hour
        std_dev_average_every_h = np.std(list(std_dev))

        max_entry = df.max(axis=0) #maximum entry for every hour
        min_entry = df.min(axis=0) #maximum entry for every hour

        # Get hour number with maximum demand
        hour_nr_max = max_entry.argmax()
        hour_nr_min = min_entry.argmin()

        # standard deviation of peak hour
        std_dev_peak_h = std_dev[hour_nr_max]

        # Difference between average and max
        diff_av_max = max_entry[hour_nr_max] - mean[hour_nr_max]
        mean_peak_h = mean[hour_nr_max]

        # Convert GW to KW
        diff_av_max = diff_av_max * 1000000 #GW to KW
        mean_peak_h = mean_peak_h * 1000000 #GW to KW

        # Weight with population
        for region_nr, n in enumerate(regions):
            if region_name == n:
                nr_of_reg = region_nr
                break
        pop = population[nr_of_reg]

        # Divide standard deviation of peak hour by population
        # which gives measure of weather variability in peak hour
        std_dev_peak_h_norm_pop = std_dev_peak_h / pop

        diff_av_max_pp = diff_av_max / pop
        mean_peak_h_pp = mean_peak_h / pop

        line_entry = [[
            str(region_name),
            mean_peak_h,
            diff_av_max,
            mean_peak_h_pp,
            diff_av_max_pp,
            std_dev_average_every_h,
            std_dev_peak_h_norm_pop]]

        line_df = pd.DataFrame(
            line_entry, columns=regional_statistics_columns)
    
        df_stats = df_stats.append(line_df)

    print(df_stats['diff_av_max'].max())
    print(df_stats['mean_peak_h'].max())
    print(df_stats['std_dev_peak_h_norm_pop'].max())
    print("-")
    print(df_stats['diff_av_max_pp'].max())
    print(df_stats['diff_av_max_pp'].min())
    print("-")
    print(df_stats['mean_peak_h_pp'].max())
    print(df_stats['mean_peak_h_pp'].min())
    # ---------------
    # Create spatial maps
    # http://darribas.org/gds15/content/labs/lab_03.html
    # http://nbviewer.jupyter.org/gist/jorisvandenbossche/57d392c085901eb4981054402b37b6b1
    # ---------------
    # Load uk shapefile
    uk_shapefile = gpd.read_file(path_shapefile)

    # Merge stats to geopanda
    shp_gdp_merged = uk_shapefile.merge(
        df_stats,
        on='name')

    # Assign projection
    crs = {'init': 'epsg:27700'} #27700: OSGB_1936_British_National_Grid
    uk_gdf = gpd.GeoDataFrame(shp_gdp_merged, crs=crs)

    ax = uk_gdf.plot()

    # Assign bin colors according to defined cmap and whether
    # plot with min_max values or only min/max values
    #bin_values = [0, 0.0025, 0.005, 0.0075, 0.01]
    #bin_values = [0, 0.02, 0.04, 0.06, 0.08, 0.1] #list(np.arange(0.0, 1.0, 0.1))

    # Field to plot
    field_to_plot = "diff_av_max_pp" # Difference between average and peak per person in KWh
    #field_to_plot = "diff_av_max"    # Difference between average and peak
    field_to_plot = 'std_dev_peak_h_norm_pop'

    nr_of_intervals = 6

    bin_values = result_mapping.get_reasonable_bin_values_II(
        data_to_plot=list(uk_gdf[field_to_plot]),
        nr_of_intervals=nr_of_intervals)
    print(float(uk_gdf[field_to_plot]))
    print("BINS " + str(bin_values))

    uk_gdf, cmap_rgb_colors, color_zero, min_value, max_value = user_defined_bin_classification(
        uk_gdf,
        field_to_plot, 
        bin_values=bin_values)

    # plot with face color attribute
    uk_gdf.plot(ax=ax, facecolor=uk_gdf['bin_color'], edgecolor='black', linewidth=0.5)

    #shp_gdp_merged.plot(column='diff_av_max', scheme='QUANTILES', k=5, cmap='OrRd', linewidth=0.1)
    #ax = uk_gdf.plot(column='diff_av_max', scheme='QUANTILES', k=5, cmap='OrRd', linewidth=0.1)
    #uk_gdf[uk_gdf['name'] == 'E06000024'].plot(ax=ax, facecolor='green', edgecolor='black')
    #uk_gdf[uk_gdf['diff_av_max'] < 0.01].plot(ax=ax, facecolor='blue', edgecolor='black')

    # Get legend patches TODO IMPROVE
    # TODO IMRPVE: MAKE CORRECT ONE FOR NEW PROCESSING
    legend_handles = result_mapping.get_legend_handles(
        bin_values[1:-1],
        cmap_rgb_colors,
        color_zero,
        min_value,
        max_value)

    plt.legend(
        handles=legend_handles,
        title="tittel_elgend",
        prop={'size': 8},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.05),
        frameon=False)

    # PLot bins on plot
    plt.text(
        0,
        -20,
        bin_values[:-1], #leave away maximum value
        fontsize=8)

    plt.tight_layout()
    plt.show()
    raise Exception
    plt.savefig(fig_name)
    plt.close()

def norm_cmap(values, cmap, vmin=None, vmax=None):
    """
    Normalize and set colormap

    Parameters
    ----------
    values : Series or array to be normalized
    cmap : matplotlib Colormap
    normalize : matplotlib.colors.Normalize
    cm : matplotlib.cm
    vmin : Minimum value of colormap. If None, uses min(values).
    vmax : Maximum value of colormap. If None, uses max(values).

    Returns
    -------
    n_cmap : mapping of normalized values to colormap (cmap)

    Source
    ------
    https://ocefpaf.github.io/python4oceanographers/blog/2015/08/24/choropleth/
    """
    mn = vmin or min(values)
    mx = vmax or max(values)
    norm = Normalize(vmin=mn, vmax=mx)
    n_cmap = plt.cm.ScalarMappable(norm=norm, cmap=cmap)

    rgb_colors = [n_cmap.to_rgba(value) for value in values]

    return n_cmap, rgb_colors

def plot_colors(rgb_colors):
    """function to plot colors
    """
    nr_dots = len(rgb_colors)

    dots = []
    x = []
    y = []
    for i in range(nr_dots):
        x.append(i + 20)
        y.append(i + 20)
    
    #plt.scatter(x, y, c=cmap, s=50)
    plt.scatter(x, y, c=rgb_colors, s=50)

    plt.show()

def user_defined_bin_classification(
        input_df,
        field_name,
        bin_values,
        cmap_diverging=None,
        cmap_sequential=None

    ):
    """Classify values according to bins

    Arguments
    ---------
    input_df : dataframe
        Dataframe to plot

    higher_as_bin : int
        Bin value of > than last bin
    cmap_sequential : str
        'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds','YlOrBr',
        'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu', 'GnBu', 'PuBu', 'YlGnBu',
        'PuBuGn', 'BuGn', 'YlGn'
    cmap_diverging : str
        'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu', 'RdYlBu', 'RdYlGn',
        'Spectral', 'coolwarm', 'bwr', 'seismic'

    Info
    -----
    Include 0 in min_max_plot == False
    Python colors:
    https://matplotlib.org/1.4.1/examples/color/colormaps_reference.html
    https://ocefpaf.github.io/python4oceanographers/blog/2015/08/24/choropleth/

    https://matplotlib.org/examples/color/colormaps_reference.html

    """
    # Check if largest value is large than last bin
    max_real_value = float(input_df[field_name].max())
    min_real_value = float(input_df[field_name].min())

    if max_real_value > 0 and min_real_value < 0:
        min_max_plot = True
    else:
        min_max_plot = False

    if not min_max_plot:

        # If only minus values
        if max_real_value < 0: #only min values
            if min_real_value > bin_values[0]:
                # add "higher as bin"
                bin_values.insert(0, min_real_value)
            elif bin_values[0] < min_real_value:
                crit_append_val = False
                raise Exception("The minimum user defined bin smaller is larger than minimum existing value")

            if not cmap_sequential:
                cmap, cmap_rgb_colors = norm_cmap(bin_values[:1], cmap='Purples') #'YlOrBr'
            else:
                cmap, cmap_rgb_colors = norm_cmap(bin_values[:1], cmap=cmap_sequential) #'YlOrBr'

        else: #only positive values
            if max_real_value > bin_values[-1]:
                # add "higher as bin"
                bin_values.append(max_real_value)
            elif bin_values[-1] > max_real_value:
                raise Exception("The maximum user defined bin value is larger than maximum min: min: {} max: {}".format(bin_values[-1], max_real_value))
            if not cmap_sequential:
                cmap, cmap_rgb_colors = norm_cmap(bin_values[1:], cmap='Purples')
            else:
                cmap, cmap_rgb_colors = norm_cmap(bin_values[1:], cmap=cmap_sequential)

        # e.g. [0, 3, 6] --> generates (0, 3], and (3, 6] bin
        input_df['bin_color'] = pd.cut(
            input_df[field_name],
            bin_values,
            include_lowest=True,
            right=True,
            labels=cmap_rgb_colors)

        color_zero = 'grey' # default
    else:

        if max_real_value < bin_values[-1]:
            raise Exception("The maximum user defined bin value is larger than maximum value {} {}".format(bin_values[-1], max_real_value))
        elif min_real_value > bin_values[0]:
            raise Exception("The minimum user defined bin smaller is larger than minimum existing value")
        else:
            pass

        # Add minimum and maximum value
        bin_values.append(max_real_value)
        bin_values.insert(0, min_real_value)

        if not cmap_diverging:
            cmap, cmap_rgb_colors = norm_cmap(bin_values, cmap='coolwarm')
        else:
            cmap, cmap_rgb_colors = norm_cmap(bin_values, cmap=cmap_diverging)

        # Reclassify zero value
        positive_bin_colors = []
        minus_bin_colors = []
        minus_bins = []
        positive_bins = [0]

        for cnt, i in enumerate(bin_values):
            if i < 0:
                minus_bin_colors.append(cmap_rgb_colors[cnt])
                minus_bins.append(i)
            elif i == 0:
                color_zero = cmap_rgb_colors[cnt]
            else:
                positive_bin_colors.append(cmap_rgb_colors[cnt])
                positive_bins.append(i)
        minus_bins.append(0)

        # ----
        # Classify
        # ----
        # Classify values in dataframe and assign color value as "bin" column
        minus_dataframe = input_df[field_name][input_df[field_name] < 0].to_frame()
        zero_dataframe = input_df[field_name][input_df[field_name] == 0].to_frame()
        plus_dataframe = input_df[field_name][input_df[field_name] > 0].to_frame()

        # e.g. [0, 3, 6] --> generates (0, 3], and (3, 6] bin
        minus_dataframe['bin_color'] = pd.cut(
            minus_dataframe[field_name],
            minus_bins,
            include_lowest=True,
            right=True,
            labels=minus_bin_colors)
        zero_dataframe['bin_color'] = [color_zero for _ in range(len(zero_dataframe))] #create list with zero color
        plus_dataframe['bin_color'] = pd.cut(
            plus_dataframe[field_name],
            positive_bins,
            include_lowest=True,
            right=True,
            labels=positive_bin_colors)
        
        # Add bins
        input_df = minus_dataframe.append(zero_dataframe)
        input_df = input_df.append(plus_dataframe)

    return input_df, cmap_rgb_colors, color_zero, min_real_value, max_real_value
    '''ax = input_df.plot()

    # Calculate color values

    #uk_gdf[uk_gdf['name'] == 'E06000024'].plot(ax=ax, facecolor='green', edgecolor='black')
    #uk_gdf[uk_gdf['diff_av_max'] < 0.01].plot(ax=ax, facecolor='blue', edgecolor='black')
    # Convert dict to dataframe
    #df = pd.DataFrame.from_dict(input_df, orient='index')

    #df['Coordinates'] = list(zip(df.longitude, df.latitude))
    #df['Coordinates'] = df['Coordinates'].apply(Point)

    # Load uk shapefile
    uk_shapefile = gpd.read_file(path_shapefile)

    # Assign correct projection
    crs = {'init': 'epsg:27700'} #27700 == OSGB_1936_British_National_Grid
    uk_gdf = gpd.GeoDataFrame(uk_shapefile, crs=crs)

    # Transform
    uk_gdf = uk_gdf.to_crs({'init' :'epsg:4326'})

    # Plot
    ax = uk_gdf.plot(color='white', edgecolor='black')

    # print coordinates
    #world.plot(column='gdp_per_cap', cmap='OrRd', scheme='quantiles');

    plt.savefig(fig_path)'''



"""testing load_factors.py
"""
from energy_demand.profiles import load_factors

import numpy as np

def test_peak_shaving_max_min():
    """Test
    """
    fuel_yh = np.zeros((2, 1, 24)) #tow fueltypes one day

    # -------
    '''fuel_yh[0][0] = [
        0.00016234, 0.00016234, 0.00016234, 0.00016234, 0.00016234, 0.00089288,
        0.00243513, 0.00389622, 0.00211045, 0.00097405, 0.00089288, 0.00105522,
        0.00089288, 0.00097405, 0.00097405, 0.00162342, 0.00243513, 0.00389622,
        0.00381504, 0.00316567, 0.00251631, 0.00235396, 0.00154225, 0.00040586]'''

    # fill all first 100 days with one
    for fueltype in range(2):
        for hour in range(12):
            fuel_yh[fueltype][0][hour] = 1
        for hour in range(12, 24):
            fuel_yh[fueltype][0][hour] = 0.5

    average_yd = np.average(fuel_yh, axis=2)
    '''loac_factor_improvement = 0.05
    loadfactor_yd_cy = load_factors.calc_lf_d(fuel_yh, average_yd, mode_constrained=False)
    loadfactor_yd_cy_improved = loadfactor_yd_cy + loac_factor_improvement'''

    loadfactor_yd_cy_improved = 0.75 + 0.1

    result = load_factors.peak_shaving_max_min(
        loadfactor_yd_cy_improved,
        average_yd,
        fuel_yh,
        mode_constrained=False)

    # ---------
    # expected
    # ---------
    max_daily_demand_allowed = 0.75 / (0.75 + 0.1)

    fuel_yh_expected = np.zeros((2, 1, 24))

    area_to_shift = (1 - max_daily_demand_allowed) * 12

    for fueltype in range(2):
        for hour in range(12):
            fuel_yh_expected[fueltype][0][hour] = max_daily_demand_allowed
        for hour in range(12, 24):
            fuel_yh_expected[fueltype][0][hour] = 0.5 + (area_to_shift / 12)

    np.testing.assert_equal(result, fuel_yh_expected)

test_peak_shaving_max_min()

def test_calc_lf_d():
    """Test
    """
    fuel_yh = np.ones((2, 2, 24)) #tow fueltype, Two day example

    fuel_yh[1][1] = np.array((range(24)))
    for i in range(12):
        fuel_yh[1][0][i] = 5
    for i in range(12, 24):
        fuel_yh[1][1][i] = 10
    average_per_day = np.average(fuel_yh, axis=2)

    result = load_factors.calc_lf_d(fuel_yh, average_per_day, mode_constrained=False)

    expected = np.zeros((2, 2))
    expected[1][0] = np.average(fuel_yh[1][0]) / np.max(fuel_yh[1][0])
    expected[1][1] = np.average(fuel_yh[1][1]) / np.max(fuel_yh[1][1])

    assert expected[1][0] == result[1][0]
    assert expected[1][1] == result[1][1]

def test_calc_lf_y():
    """Test
    """
    # fueltype, days, hours
    fuel_yh = np.ones((8, 365, 24)) #Two day example
    fuel_yh[2][1] = np.array((range(24)))
    for i in range(12):
        fuel_yh[2][0][i] = 5
    for i in range(12, 24):
        fuel_yh[2][1][i] = 10

    result = load_factors.calc_lf_y(fuel_yh)

    expected = np.zeros((8))
    expected[0] = np.average(fuel_yh[0]) / np.max(fuel_yh[0])
    expected[2] = np.average(fuel_yh[2]) / np.max(fuel_yh[2])

    assert expected[0] == result[0]
    assert expected[2] == result[2]

def test_calc_lf_season():
    """Test
    """
    # fueltype, days, hours
    fuel_yh = np.ones((8, 3, 24))
    seasons = {'seasonA': [1, 2]}  #Two day example, seasonA is second and third day

    fuel_yh[2][1] = np.array((range(24)))
    for i in range(12):
        fuel_yh[2][0][i] = 5
    for i in range(12, 24):
        fuel_yh[2][1][i] = 10
    average_per_day = np.average(fuel_yh, axis=2)

    result = load_factors.calc_lf_season(seasons, fuel_yh, average_per_day)

    expected = {'seasonA': np.zeros((8))}
    expected['seasonA'][2] = np.average(fuel_yh[2]) / np.max(fuel_yh[2][1:3])

    #if within seasons
    #expected['seasonA'][2] = np.average(fuel_yh[2][1:3]) / np.max(fuel_yh[2][1:3]) * 100
    assert expected['seasonA'][2] == result['seasonA'][2]



"""Generate model runs
"""
import os

from energy_demand.read_write import write_data

def generate_yaml(
        narrative,
        params,
        out_folder,
        weather_real,
        name_sector_model
    ):

    # Create narratives
    narratives = []

    for param_name in params:
        #Narratives
        narratives.append({
            'name': param_name,
            'provides': {'energy_demand_constrained': [param_name]},
            'description': '',
            'variants': [
                {
                    'name': narrative,
                    'description': '',
                    'data': [{param_name: '{}__{}.csv'.format(param_name, narrative)}]
                }]
        })


    yaml_dict = {

        'name': name_sector_model,

        #Narratives
        'narratives': narratives,

        # Scenarios
        'scenarios': ['population', 'gva_per_head','temperature'],

        'description': 'Model run description is...',

        'model_dependencies': [],

        'sector_models': ['energy_demand_constrained'],

        #Scenario dependendies
        'scenario_dependencies': [
            {
                'sink': name_sector_model,
                'source': 'population',
                'source_output': 'population',
                'sink_input': 'population'
            },
            {
                'sink': name_sector_model,
                'source': 'gva_per_head',
                'source_output': 'gva_per_head',
                'sink_input': 'gva_per_head'
            },
            {
                'sink': name_sector_model,
                'source': 'gva_per_sector',
                'source_output': 'gva_per_sector',
                'sink_input': 'gva_per_sector'
            },
            {
                'sink': name_sector_model,
                'source': 'weather_station_coordinates',
                'source_output': 'latitude',
                'sink_input': 'latitude'
            },
            {
                'sink': name_sector_model,
                'source': 'weather_station_coordinates',
                'source_output': 'longitude',
                'sink_input': 'longitude'
            },
            {
                'sink': name_sector_model,
                'source': 'temperature',
                'source_output': 't_min',
                'sink_input': 't_min'
            },
            {
                'sink': name_sector_model,
                'source': 'temperature',
                'source_output': 't_max',
                'sink_input': 't_max'
            }],


        }

    path_yaml = os.path.join(out_folder, "energy_demand_constrained__{}.yml".format(weather_real))
    write_data.write_yaml(yaml_dict, path_yaml)

params = [
    'air_leakage',
    'assump_diff_floorarea_pp',
    'cooled_floorarea',
    'dm_improvement',
    'f_eff_achieved',
    'generic_enduse_change',
    'heat_recovered',
    'is_t_base_heating',
    'p_cold_rolling_steel',
    'rs_t_base_heating',
    'ss_t_base_heating',
    'smart_meter_p',
    'generic_fuel_switch']

out_folder = "C:/_scrap"

#Weather scenario
weather_real = "NF1"
name_sector_model = "energy_demand_constrained"

generate_yaml(
    narrative='central',
    params=params,
    out_folder=out_folder,
    weather_real=weather_real,
    name_sector_model=name_sector_model)



"""
"""
from energy_demand.initalisations import helpers
from energy_demand.read_write import read_data

def test_init_dict_brackets(): 
    """Test
    """
    first_level_keys = ["a", "b"]
    one_level_dict = helpers.init_dict_brackets(first_level_keys)
    expected = {"a": {}, "b": {}}

    assert one_level_dict == expected

def test_get_nested_dict_key():
    """Test
    """
    nested_dict = {"a": {"c": 1, "d": 3}, "b": {"e": 4}}
    keys = helpers.get_nested_dict_key(nested_dict)

    assert "c" in keys
    assert "d" in keys
    assert "e" in keys

def test_set_same_eff_all_tech():
    """Test
    """
    eff_to_assign = 0.5

    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0)}

    techs_eff = helpers.set_same_eff_all_tech(
        technologies=technologies,
        f_eff_achieved=0.44)

    assert techs_eff['boilerA'].eff_achieved == 0.44



"""Script to fit technology diffusion

This script calculates the three parameters of a sigmoid diffusion
for every technology which is diffused and has a larger service
fraction at the model end year
"""
from collections import defaultdict
import numpy as np
from scipy.optimize import curve_fit
import pylab
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg') # Used to make it work in linux

from energy_demand.technologies import diffusion_technologies


def plotout_sigmoid_tech_diff(
        L_value,
        technology,
        xdata,
        ydata,
        fit_parameter,
        plot_crit=False,
        close_window_crit=True
    ):
    """Plot sigmoid diffusion
    """
    def close_event():
        """Timer to close window automatically
        """
        plt.close()

    x = np.linspace(1990, 2110, 300)
    y = diffusion_technologies.sigmoid_function(x, L_value, *fit_parameter)

    fig = plt.figure()

    #creating a timer object and setting an interval
    timer = fig.canvas.new_timer(interval=555)
    timer.add_callback(close_event)

    fig.set_size_inches(12, 8)
    pylab.plot(xdata, ydata, 'o', label='base year and future market share')
    pylab.plot(x, y, label='fit')

    pylab.ylim(0, 1.05)
    pylab.legend(loc='best')

    pylab.xlabel('Time')
    pylab.ylabel('Market share of technology on energy service')
    pylab.title("Sigmoid diffusion of technology {}".format(technology))

    if plot_crit:
        if close_window_crit:
            pylab.show()
        else:
            timer.start()
            pylab.show()
            pass
    else:
        pass

def calc_sigmoid_parameters(
        l_value,
        xdata,
        ydata,
        fit_assump_init=0.001,
        error_range=0.00002,
        number_of_iterations=1000
    ):
    """Calculate sigmoid parameters. Check if fitting is good enough.

    Arguments
    ----------
    l_value : float
        Maximum upper level
    xdata : array
        X data
    ydata : array
        Y data
    fit_assump_init : float
        Small value for correct sigmoid diffusion in case start value is equal to L
    fit_crit_max : float
        Criteria to control and abort fit
    fit_crit_min : float
        Criteria to control and abort fit (slope must be posititive)
    error_range : float,default=0.00002
        Allowed fitting offset in percent

    Note
    -------
        `error_range` can be changed if the plotting is weird. If you increase
        chances are however higher that the fitting does not work anymore.

    Returns
    ------
    fit_parameter : array
        Parameters (first position: midpoint, second position: slope)
    """
    # ---------------------------------------------
    # Generate possible starting parameters for fit
    # ---------------------------------------------
    start_param_list = [0.0, 1.0, 0.0001, 0.001, 0.01]

    # ---------------------------------------------
    # Fit
    # ---------------------------------------------
    cnt = 0
    successfull = False
    while not successfull:
        try:
            start_parameters = [
                round(start_param_list[cnt], 3),
                round(start_param_list[cnt], 3)]

            # ------------------------------------------------
            # Test if parameter[1] should be minus or positive
            # ------------------------------------------------
            if ydata[0] < ydata[1]: # end point has higher value
                crit_plus_minus = 'plus'
            else:
                crit_plus_minus = 'minus'

                if l_value == ydata[0]:
                    l_value += fit_assump_init

            # Select start parameters depending on pos or neg diff
            if crit_plus_minus == 'minus':
                start_parameters[0] *= 1
                start_parameters[1] *= -1

            #logging.debug(" patameters {} {} {} {}".format(
            #    xdata, ydata, start_parameters, l_value))

            # Fit function
            fit_parameter = fit_sigmoid_diffusion(
                l_value,
                xdata,
                ydata,
                start_parameters,
                number_of_iterations=number_of_iterations)

            # Test if start paramters are identical to fitting parameters
            if (crit_plus_minus == 'plus' and fit_parameter[1] < 0) or (
                    crit_plus_minus == 'minus' and fit_parameter[1] > 0):
                cnt += 1
                if cnt >= len(start_param_list):
                    raise Exception("Error: Sigmoid curve fitting failed")
            else:
                if (fit_parameter[0] == start_parameters[0]) or (
                        fit_parameter[1] == start_parameters[1]):
                    cnt += 1
                    if cnt >= len(start_param_list):
                        raise Exception("Error: Sigmoid curve fitting failed")
                else:
                    # Check how good the fit is
                    y_calculated_ey = diffusion_technologies.sigmoid_function(
                        x_value=xdata[1],
                        l_value=l_value,
                        midpoint=fit_parameter[0],
                        steepness=fit_parameter[1])

                    y_calculated_by = diffusion_technologies.sigmoid_function(
                        x_value=xdata[0],
                        l_value=l_value,
                        midpoint=fit_parameter[0],
                        steepness=fit_parameter[1])

                    fit_measure_p_by = float((100.0 / ydata[0]) * y_calculated_by)
                    fit_measure_p_ey = float((100.0 / ydata[1]) * y_calculated_ey)

                    if (fit_measure_p_ey < (100.0 - error_range) or fit_measure_p_ey > (100.0 + error_range)) or (
                        fit_measure_p_by < (100.0 - error_range) or fit_measure_p_by > (100.0 + error_range)):
                        
                        #logging.debug("... Fitting measure %s %s (percent) is not good enough", fit_measure_p_by, fit_measure_p_ey)
                        successfull = False
                        cnt += 1
                    else:
                        successfull = True
                        '''plotout_sigmoid_tech_diff(
                            l_value,
                            "FINISHED FITTING",
                            xdata,
                            ydata,
                            fit_parameter,
                            plot_crit=True,
                            close_window_crit=True)'''
        except (RuntimeError, IndexError):
            cnt += 1

            if cnt >= len(start_param_list):
                raise Exception(
                    "Fitting did not work: Check whether start year is <= the year 2000")

    return fit_parameter

def fit_sigmoid_diffusion(
        l_value,
        x_data,
        y_data,
        start_parameters,
        number_of_iterations=10000
    ):
    """Fit sigmoid curve based on two points on the diffusion curve

    Arguments
    ----------
    l_value : float
        The sigmoids curve maximum value (max consumption)
    x_data : array
        X coordinate of two points
    y_data : array
        X coordinate of two points
    number_of_iterations : int
        Number of iterations used for sigmoid fitting

    Returns
    -------
    popt : dict
        Fitting parameters (l_value, midpoint, steepness)

    Note
    ----
    The Sigmoid is substacted - 2000 to allow for better fit with low values

    Warning
    -------
    It cannot fit a value starting from 0. Therefore, some initial penetration needs
    to be assumed (e.g. 0.001%)
    RuntimeWarning is ignored
    https://stackoverflow.com/questions/4359959/overflow-in-exp-in-scipy-numpy-in-python
    """
    def sigmoid_fitting_function(x_value, x0_value, k_value):
        """Sigmoid function used for fitting
        """
        with np.errstate(over='ignore'): #RuntimeWarning: overflow encountered in exp
            y_value = l_value / (1 + np.exp(-k_value * ((x_value - 2000.0) - x0_value)))
            return y_value

    popt, _ = curve_fit(
        sigmoid_fitting_function,
        x_data,
        y_data,
        p0=start_parameters,
        maxfev=number_of_iterations)

    return popt

def tech_l_sigmoid(
        s_tech_switched_ey,
        enduse_fuel_switches,
        technologies,
        installed_tech,
        s_fueltype_by_p,
        s_tech_by_p,
        fuel_tech_p_by
    ):
    """Calculate L value (maximum possible theoretical market penetration)
    for every installed technology with maximum theoretical replacement
    share (calculate second sigmoid point).

    Arguments
    ----------
    s_tech_switched_ey : dict
        Ey service tech
    enduse_fuel_switches : dict
        Fuel switches of enduse
    installed_tech : dict
        Installed technologies (as keys)
    s_fueltype_by_p : dict
        Service share per fueltye of base year
    s_tech_by_p : dict
        Percentage of service demands for every technology
    fuel_tech_p_by : dict
        Fuel share per technology of base year

    Returns
    -------
    l_values_sig : dict
        L value for sigmoid diffusion of all technologies
        for which a switch is implemented
    """
    l_values_sig = {}

    # Check wheter there are technologies in this enduse which are switched
    if installed_tech == []:
        pass
    else:
        # Iterite list with enduses where fuel switches are defined
        for technology in installed_tech:

            # If decreasing technology, L-Value stays initial value
            if s_tech_by_p[technology] > s_tech_switched_ey[technology]:
                l_values_sig[technology] = s_tech_by_p[technology]

            else:
                # Calculate maximum service demand for specific tech
                tech_install_p = calc_service_fuel_switched(
                    enduse_fuel_switches,
                    technologies,
                    s_fueltype_by_p,
                    s_tech_by_p,
                    fuel_tech_p_by,
                    'max_switch')

                # Read L-values with calculating maximum sigmoid theoretical diffusion
                l_values_sig[technology] = tech_install_p[technology]

    return l_values_sig

def calc_service_fuel_switched(
        fuel_switches,
        technologies,
        s_fueltype_by_p,
        s_tech_by_p,
        fuel_tech_p_by,
        switch_type
    ):
    """Calculate energy service demand percentages after fuel switches.

    Arguments
    ----------
    fuel_switches : dict
        Fuel switches of a specific enduse
    technologies : dict
        Technologies
    s_fueltype_by_p : dict
        Service share per fueltype in base year
    s_tech_by_p : dict
        Share of service demand per technology for base year
    fuel_tech_p_by : dict
        Fuel shares for each technology of an enduse for base year
    switch_type : str
        If either switch is for maximum theoretical switch
        of with defined switch until end year

    Return
    ------
    s_tech_switched_p : dict
        Service in future year with added and substracted
        service demand for every technology

    Note
    ----
    Assertion may be removed to increase speed
    """
    s_tech_switched_p = {}

    for fuel_switch in fuel_switches:

        tech_install = fuel_switch.technology_install
        tech_replace_fueltype = fuel_switch.fueltype_replace

        # Share of energy service of repalced fueltype before switch in base year
        service_p_by = s_fueltype_by_p[tech_replace_fueltype]

        # Service share of fueltype that will be switched
        if switch_type == 'max_switch':
            s_diff_fueltype_by_p = service_p_by * technologies[tech_install].tech_max_share
        elif switch_type == 'actual_switch':
            s_diff_fueltype_by_p = service_p_by * fuel_switch.fuel_share_switched_ey

        # ----------------
        # Service addition
        # ----------------
        s_tech_switched_p[tech_install] = s_tech_by_p[tech_install] + s_diff_fueltype_by_p

        # ----------------
        # Service substraction
        # ----------------
        # Iterate technologies which are replaced for this fueltype and substract service demand proportionally

        # Technologies with lower demands
        technologies_replaced = list(fuel_tech_p_by[tech_replace_fueltype].keys())

        # Calculate proportional share of technologies in replaced fueltype in by
        tot_by_share = 0
        for tech in technologies_replaced:
            tot_by_share += s_tech_by_p[tech]

        # Substract switched share proportionally to base year service in fueltype
        for tech in technologies_replaced:

            # Because of rounding error otherwise minus values possible
            round_digits = 5

            # Relative share in by
            rel_tech_by_p = s_tech_by_p[tech] / tot_by_share

            # Substract (service_by - service to switch * relative share)
            s_tech_switched_p[tech] = round(s_tech_by_p[tech], round_digits) - round(s_diff_fueltype_by_p * rel_tech_by_p, round_digits)

            assert s_tech_switched_p[tech] >= 0

    # -----------------------
    # Calculate service fraction of all technologies in
    # enduse not affected by fuel switch
    # -----------------------
    s_affected_p_ey = sum(s_tech_switched_p.values())
    unaffected_service_to_distr_p = 1 - s_affected_p_ey

    # Calculate service fraction of remaining technologies
    fractions_unaffected_switch = {}
    for tech, s_tech_p in s_tech_by_p.items():
        if tech not in s_tech_switched_p.keys():
            fractions_unaffected_switch[tech] = s_tech_p

    # Sum of unaffected service shares
    service_tot_remaining = sum(fractions_unaffected_switch.values())

    # Get relative distribution of all not affected techs
    for tech, tech_fraction in fractions_unaffected_switch.items():

        # Relative share
        s_rel_fraction_p = tech_fraction / service_tot_remaining
        s_tech_switched_p[tech] = s_rel_fraction_p * unaffected_service_to_distr_p

        assert s_tech_switched_p[tech] >= 0

    return dict(s_tech_switched_p)

def get_tech_installed(enduse, fuel_switches):
    """Read out all technologies which are specifically switched
    to of a specific enduse

    Parameter
    ---------
    enduse : str
        enduse
    fuel_switches : dict
        All fuel switches where a share of a fuel
        of an enduse is switched to a specific technology

    Return
    ------
    installed_tech : list
        List with all technologies where a fuel share is switched to
    crit_fuel_switch : bool
        Criteria wheter swich is defined or not
    """
    # Add technology list for every enduse with affected switches
    installed_tech = set([])

    for switch in fuel_switches:
        if switch.enduse == enduse:
            installed_tech.add(switch.technology_install)

    if len(list(installed_tech)) > 0:
        crit_fuel_switch = True
    else:
        crit_fuel_switch = False

    return list(installed_tech), crit_fuel_switch

def get_l_values(
        technologies,
        technologies_to_consider,
        regions=False
    ):
    """Get l values (Maximum shares of each technology)
    for all installed technologies

    Arguments
    ----------
    technologies : dict
        Technologies
    technologies_to_consider : list
        Technologies to consider
    s_fueltype_by_p :
        Fraction of service per fueltype in base year
    regions : dict
        Regions

    Return
    ------
    l_values_sig : dict
        Sigmoid paramters
    """
    l_values_sig = defaultdict(dict)

    for region in regions:
        if technologies_to_consider == []:
            pass
        else:
            for tech in technologies_to_consider:
                l_values_sig[region][tech] = technologies[tech].tech_max_share

    return dict(l_values_sig)

def tech_sigmoid_parameters(
        yr_until_switched,
        switch_yr_start,
        technologies,
        l_values,
        s_tech_by_p,
        s_tech_switched_p,
        fit_assump_init=0.001,
        plot_sigmoid_diffusion=False
    ):
    """Calculate sigmoid diffusion parameters based on energy service
    demand in base year and projected future energy service demand. The
    future energy servie demand is calculated based on fuel switches.

    Three potential sigmoid outputs are possible:

        'linear':       No sigmoid fitting possible because the
                        service in the future year is identical
                        to the service in the base year

        None:           No sigmoid is fitted if the future
                        service share is zero.

        fit_parameters  Sigmoid diffusion parameters

    Because the sigmoid fitting does not work if the initial and
    end values are zero, small approximatie values `fit_assump_init`
    are inserted to allow the function 'calc_sigmoid_parameters'
    to fit.

    Arguments
    ----------
    yr_until_switched : int
        Year until switch is fully realised
    switch_yr_start : int
        Start year of story in narrative
    technologies : dict
        technologies
    l_values : dict
        L values for maximum possible diffusion of technologies
    s_tech_by_p : dict
        Energy service demand for base year (1. sigmoid point)
    s_tech_switched_p : dict
        Service demand after fuelswitch
    fit_assump_init : float
        Approximation helping small number to allow fit
    plot_sigmoid_diffusion : bool,default=True
        Criteria whether sigmoid are plotted

    Returns
    -------
    sig_params : dict
        Sigmoid diffusion parameters to read energy service demand percentage (not fuel!)

    Info
    -----
        `rounding_accuracy`     This rounds the by and ey service share
                                to the defined number of digits. Because of
                                rounding error, there might be very small
                                differences in percentual service demand.
    """
    def calc_m(x1, x2, y1, y2):
        m = (y1-y2) / (x1 - x2)
        return m

    def calc_c(m, x1, y1):
        c = y1 - (m * x1)
        return c

    rounding_accuracy = 4       # Criteria how much difference in % can be rounded
    linear_approx_crit = 0.001  # Criteria to simplify with linear approximation if difference is smaller (decimal)
    error_range = 0.0002        # Error how close the fit must be
    number_of_iterations = 100  # Number of iterations of sigmoid fitting algorithm

    # Technologies to apply calculation
    installed_techs = s_tech_by_p.keys()
    sig_params = defaultdict(dict)

    # Fitting criteria where the calculated sigmoid slope and midpoint can be provided limits
    if installed_techs == []:
        pass
    else:
        for tech in installed_techs:

            # Test whether technology has the market entry before or after base year,
            # If afterwards, set very small number in market entry year
            if technologies[tech].market_entry > switch_yr_start:
                point_x_by = technologies[tech].market_entry
                point_y_by = fit_assump_init
            else:
                point_x_by = switch_yr_start   # Base year
                point_y_by = s_tech_by_p[tech] # Base year service share

                # If the base year is the market entry year use a very small number
                if point_y_by == 0:
                    point_y_by = fit_assump_init
                #elif point_y_by == 1.0:
                #    point_y_by = 1 - fit_assump_init

            # Future energy service demand
            point_x_ey = yr_until_switched
            point_y_ey = s_tech_switched_p[tech]

            # If future share is zero, entry small value
            if point_y_ey == 0:
                point_y_ey = fit_assump_init
            elif point_y_ey == 1.0:
                point_y_ey = 1 - fit_assump_init
            else:
                pass

            # Data of the two points
            xdata = np.array([point_x_by, point_x_ey])
            ydata = np.array([point_y_by, point_y_ey])

            # If no change in by to ey but not zero (linear change)
            if (round(point_y_by, rounding_accuracy) == round(point_y_ey, rounding_accuracy)) and (
                    point_y_ey != fit_assump_init) and (
                        point_y_by != fit_assump_init):
            #if 1 == 1:
                # Linear diffusion (because by and ey share are identical)
                sig_params[tech]['midpoint'] = 'linear'
                sig_params[tech]['steepness'] = 'linear'
                sig_params[tech]['l_parameter'] = 'linear'

                # Calculate linear slope and linear y-intercept (with two data points)
                sig_params[tech]['linear_slope'] = calc_m(xdata[0], xdata[1], ydata[0], ydata[1])
                sig_params[tech]['linear_y_intercept'] = calc_c(sig_params[tech]['linear_slope'], xdata[0], ydata[0])
            else:
                # Test if no increase or decrease or if no future potential share
                if (point_y_by == fit_assump_init and point_y_ey == fit_assump_init) or (l_values[tech] == 0):
                    sig_params[tech]['midpoint'] = None
                    sig_params[tech]['steepness'] = None
                    sig_params[tech]['l_parameter'] = None
                else:

                    # If difference is smaller than a certain share, approximate with linear
                    if abs(ydata[1] - ydata[0]) < linear_approx_crit:
                        sig_params[tech]['midpoint'] = 'linear'
                        sig_params[tech]['steepness'] = 'linear'
                        sig_params[tech]['l_parameter'] = 'linear'

                        # Calculate linear slope and linear y-intercept (with two data points)
                        sig_params[tech]['linear_slope'] = calc_m(xdata[0], xdata[1], ydata[0], ydata[1])
                        sig_params[tech]['linear_y_intercept'] = calc_c(sig_params[tech]['linear_slope'], xdata[0], ydata[0])

                    try:
                        # Parameter fitting
                        fit_parameter = calc_sigmoid_parameters(
                            l_values[tech],
                            xdata,
                            ydata,
                            fit_assump_init=fit_assump_init,
                            error_range=error_range,
                            number_of_iterations=number_of_iterations)

                        sig_params[tech]['midpoint'] = fit_parameter[0]
                        sig_params[tech]['steepness'] = fit_parameter[1]
                        sig_params[tech]['l_parameter'] = l_values[tech]
                    except:
                        """If sigmoid fitting failed, implement linear diffusion

                        The sigmoid diffusion may fail if the fitting does not work
                        because the points to fit are too similar. This could be 
                        improved by increasing the number of iterations 'number_of_iterations'
                        at higher computation costs
                        """
                        #logging.warning("Instead of sigmoid a linear approximation is used %s %s %s", tech, xdata, ydata)
                        sig_params[tech]['midpoint'] = 'linear'
                        sig_params[tech]['steepness'] = 'linear'
                        sig_params[tech]['l_parameter'] = 'linear'

                        # Calculate linear slope and linear y-intercept (with two data points)
                        sig_params[tech]['linear_slope'] = calc_m(xdata[0], xdata[1], ydata[0], ydata[1])
                        sig_params[tech]['linear_y_intercept'] = calc_c(sig_params[tech]['linear_slope'], xdata[0], ydata[0])

    return dict(sig_params)



"""Functions which are writing data
"""
import os
import logging
import configparser
import csv
import numpy as np
from ruamel.yaml import YAML

from energy_demand.basic import lookup_tables, date_prop, basic_functions, conversions
from energy_demand import enduse_func


def write_yaml(data, file_path):
    """Write plain data to a file as yaml

    Parameters
    ----------
    data
        Data to write (should be lists, dicts and simple values)
    file_path : str
        The path of the configuration file to write
    """
    with open(file_path, 'w+') as file_handle:
        yaml = YAML(typ='unsafe')
        yaml.default_flow_style = False
        yaml.allow_unicode = True
        return yaml.dump(data, file_handle)


def logg_info(modelrun, fuels_in, data):
    """Logg information
    """
    lookups = lookup_tables.basic_lookups()
    logging.info("=====================================================")
    logging.info("Simulation year:         %s", str(modelrun.curr_yr))
    logging.info("Nr of regions:           %s", str(data['assumptions'].reg_nrs))
    logging.info("Total ktoe:              %s", str(conversions.gwh_to_ktoe(fuels_in["fuel_in"])))
    logging.info("-----------------------------------------------------")
    logging.info("[GWh] Total input:       %s", str(fuels_in["fuel_in"]))
    logging.info("[GWh] Total output:      %s", str(np.sum(modelrun.ed_fueltype_national_yh)))
    logging.info("[GWh] Total difference:  %s", str(round((np.sum(modelrun.ed_fueltype_national_yh) - fuels_in["fuel_in"]), 4)))
    logging.info("-----------------------------------------------------")
    logging.info("[GWh] oil input:         %s", str(fuels_in["fuel_in_oil"]))
    logging.info("[GWh] oil output:        %s", str(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['oil']])))
    logging.info("[GWh] oil diff:          %s", str(round(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['oil']]) - fuels_in["fuel_in_oil"], 4)))
    logging.info("-----------------------------------------------------")
    logging.info("[GWh] biomass output:    %s", str(fuels_in["fuel_in_biomass"]))
    logging.info("[GWh] biomass output:    %s", str(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['biomass']])))
    logging.info("[GWh] biomass diff:      %s", str(round(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['biomass']]) - fuels_in["fuel_in_biomass"], 4)))
    logging.info("-----------------------------------------------------")
    logging.info("[GWh] solid_fuel output: %s", str(fuels_in["fuel_in_solid_fuel"]))
    logging.info("[GWh] solid_fuel output: %s", str(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['solid_fuel']])))
    logging.info("[GWh] solid_fuel diff:   %s", str(round(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['solid_fuel']]) - fuels_in["fuel_in_solid_fuel"], 4)))
    logging.info("-----------------------------------------------------")
    logging.info("[GWh] elec output:       %s", str(fuels_in["fuel_in_elec"]))
    logging.info("[GWh] elec output:       %s", str(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['electricity']])))
    logging.info("[GWh] ele fuel diff:     %s", str(round(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['electricity']]) - fuels_in["fuel_in_elec"], 4)))
    logging.info("-----------------------------------------------------")
    logging.info("[GWh] gas output:        %s", str(fuels_in["fuel_in_gas"]))
    logging.info("[GWh] gas output:        %s", str(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['gas']])))
    logging.info("[GWh] gas diff:          %s", str(round(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['gas']]) - fuels_in["fuel_in_gas"], 4)))
    logging.info("-----------------------------------------------------")
    logging.info("[GWh] hydro output:      %s", str(fuels_in["fuel_in_hydrogen"]))
    logging.info("[GWh] hydro output:      %s", str(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['hydrogen']])))
    logging.info("[GWh] hydro diff:        %s", str(round(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['hydrogen']]) - fuels_in["fuel_in_hydrogen"], 4)))
    logging.info("-----------------------------------------------------")
    logging.info("TOTAL HEATING            %s", str(fuels_in["tot_heating"]))
    logging.info("[GWh] heat input:        %s", str(fuels_in["fuel_in_heat"]))
    logging.info("[GWh] heat output:       %s", str(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['heat']])))
    logging.info("[GWh] heat diff:         %s", str(round(np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['heat']]) - fuels_in["fuel_in_heat"], 4)))
    logging.info("-----------------------------------------------------")
    logging.info("Diff elec p:             %s", str(round((np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['electricity']])/ fuels_in["fuel_in_elec"]), 4)))
    logging.info("Diff gas p:              %s", str(round((np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['gas']])/ fuels_in["fuel_in_gas"]), 4)))
    logging.info("Diff oil p:              %s", str(round((np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['oil']])/ fuels_in["fuel_in_oil"]), 4)))
    logging.info("Diff solid_fuel p:       %s", str(round((np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['solid_fuel']])/ fuels_in["fuel_in_solid_fuel"]), 4)))
    logging.info("Diff hydrogen p:         %s", str(round((np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['hydrogen']])/ fuels_in["fuel_in_hydrogen"]), 4)))
    logging.info("Diff biomass p:          %s", str(round((np.sum(modelrun.ed_fueltype_national_yh[lookups['fueltypes']['biomass']])/ fuels_in["fuel_in_biomass"]), 4)))
    logging.info("=====================================================")

def tuple_representer(dumper, data):
    return dumper.represent_scalar(
        tag=u'tag:yaml.org,2002:str',
        value='({}, {})'.format(data[0], data[1]))

def write_array_to_txt(path_result, array):
    """Write scenario population for a year to txt file
    """
    np.savetxt(path_result, array, delimiter=',')

def write_list_to_txt(path_result, list_out):
    """Write scenario population for a year to txt file
    """
    file = open(path_result, "w")
    for entry in list_out:
        file.write(entry + "\n")

def write_scenaric_population_data(sim_yr, path_result, pop_y):
    """Write scenario population for a year to '.npy' file

    Parameters
    ----------
    sim_yr : int
        Simulation year
    path_result : str
        Path to resulting folder
    pop_y : array or list
        Population of simulation year
    """
    path_file = os.path.join(
        path_result,
        "pop__{}__{}".format(sim_yr, ".npy"))

    np.save(path_file, pop_y)
    assert type(pop_y) != dict #test that not a dict

    logging.debug("... finished saving population")

def dump(data, file_path):
    """Write plain data to a file as yaml

    Parameters
    ----------
    file_path : str
        The path of the configuration file to write
    data
        Data to write (should be lists, dicts and simple values)
    """
    write_yaml(data, file_path)

def write_yaml_output_keynames(path_yaml, key_names):
    """Generate YAML file where the outputs
    for the sector model can be easily copied

    Arguments
    ----------
    path_yaml : str
        Path where yaml file is saved
    key_names : dict
        Names of keys of supply_out dict
    """
    list_to_dump = []

    for key_name in key_names:
        dict_to_dump = {
            'name': key_name,
            'spatial_resolution': 'lad_uk_2016',
            'temporal_resolution': 'hourly',
            'units': 'GWh'}

        list_to_dump.append(dict_to_dump)

    dump(list_to_dump, path_yaml)

def write_yaml_param_scenario(path_yaml, dict_to_dump):
    """Write all strategy variables to YAML file

    Arguments
    ----------
    path_yaml : str
        Path where yaml file is saved
    dict_to_dump : dict
        Dict which is written to YAML
    """
    list_to_dump = [dict_to_dump]
    dump(list_to_dump, path_yaml)

def write_yaml_param_complete(path_yaml, dict_to_dump):
    """Write all strategy variables to YAML file

    Arguments
    ----------
    path_yaml : str
        Path where yaml file is saved
    dict_to_dump : dict
        Dict which is written to YAML
    """
    list_to_dump = []

    for paramter_info in dict_to_dump:
        dump_dict = {}
        dump_dict['suggested_range'] = paramter_info['suggested_range']
        dump_dict['absolute_range'] = paramter_info['absolute_range']
        dump_dict['description'] = paramter_info['description']
        dump_dict['name'] = paramter_info['name']
        dump_dict['default_value'] = paramter_info['default_value']
        dump_dict['units'] = paramter_info['units']

        list_to_dump.append(dump_dict)

    # Dump list
    dump(list_to_dump, path_yaml)

def write_simulation_inifile(path, data, simulated_regions):
    """Create .ini file with simulation parameters which are
    used to read in correctly the simulation results

    Arguments
    ---------
    paths: str
        Path
    data : dict
        Data container

    simulated_regions : list
        Simulated regions
    """
    path_ini_file = os.path.join(
        path, 'model_run_sim_param.ini')

    config = configparser.ConfigParser()

    config.add_section('SIM_PARAM')
    config['SIM_PARAM']['reg_nrs'] = str(data['assumptions'].reg_nrs)
    config['SIM_PARAM']['base_yr'] = str(data['assumptions'].base_yr)
    config['SIM_PARAM']['sim_yrs'] = str(data['assumptions'].sim_yrs)

    # ----------------------------
    # Other information to pass to plotting and summing function
    # ----------------------------
    config.add_section('ENDUSES')

    #convert list to strings
    config['ENDUSES']['residential'] = str(data['enduses']['residential'])
    config['ENDUSES']['service'] = str(data['enduses']['service'])
    config['ENDUSES']['industry'] = str(data['enduses']['industry'])

    config.add_section('REGIONS')
    config['REGIONS']['regions'] = str(simulated_regions)

    with open(path_ini_file, 'w') as write_info:
        config.write(write_info)

def write_min_max_result_to_txt(file_path, values, yearday, yearday_date):
    """Write min and max values including data information to csv file
    """
    file = open(file_path, "w")
    file.write("Yearday: {}".format(yearday) + '\n')
    file.write("date " + str(yearday_date) + '\n')
    for value in values:
        file.write("{}".format(value) + '\n') #Write list to values
    file.close()

    return

def write_lf(
        path_result_folder,
        path_new_folder,
        parameters,
        model_results,
        file_name
    ):
    """Write numpy array to `.npy` file

    path_result_folder,
    path_new_folder,
    parameters,
    model_results,
    file_name
    """
    # Create folder and subolder
    basic_functions.create_folder(
        path_result_folder)

    path_result_sub_folder = os.path.join(
        path_result_folder, path_new_folder)

    basic_functions.create_folder(path_result_sub_folder)

    # Create full file_name
    for name_param in parameters:
        file_name += str("__") + str(name_param)

    path_file_fueltype = os.path.join(path_result_sub_folder, file_name) + "__" + ".npy"

    np.save(path_file_fueltype, model_results)

def write_only_peak_total_regional(
        sim_yr,
        name_new_folder,
        path_result,
        model_results,
        file_name_annual_sum
        ):
    """Write only total regional demand for a region
    """
    path_result_sub_folder = os.path.join(
        path_result, name_new_folder)

    basic_functions.create_folder(
        path_result_sub_folder)

    path_file_annual_sum = os.path.join(
        path_result_sub_folder,
        "{}__{}{}".format(file_name_annual_sum, sim_yr, ".npy"))

    # ------------------------------------
    # Sum annual fuel across all fueltypes
    # ------------------------------------
    # Sum across 8760 hours
    ed_fueltype_regs_y = np.sum(model_results, axis=2)
    np.save(path_file_annual_sum, ed_fueltype_regs_y)

def write_only_peak(
        sim_yr,
        name_new_folder,
        path_result,
        model_results,
        file_name_peak_day
    ):
    """Write only peak demand and total regional demand for a region
    """
    path_result_sub_folder = os.path.join(
        path_result, name_new_folder)

    basic_functions.create_folder(
        path_result_sub_folder)

    path_file_peak_day = os.path.join(
        path_result_sub_folder,
        "{}__{}{}".format(file_name_peak_day, sim_yr, ".npy"))

    # ------------------------------------
    # Write out peak electricity day demands
    # ------------------------------------
    # Get peak day electricity
    lookups = lookup_tables.basic_lookups()
    fueltype_int = lookups['fueltypes']['electricity']

    national_hourly_demand = np.sum(model_results[fueltype_int], axis=0)
    peak_day_electricity, _ = enduse_func.get_peak_day_single_fueltype(national_hourly_demand)
    selected_hours = date_prop.convert_yearday_to_8760h_selection(peak_day_electricity)
    selected_demand = model_results[:, :, selected_hours]

    np.save(path_file_peak_day, selected_demand)

def write_fueltype_reg_8760(
        sim_yr,
        name_new_folder,
        path_result,
        model_results,
        file_name_peak_day
    ):
    """Write only peak demand and total regional demand for a region
    """
    path_result_sub_folder = os.path.join(
        path_result, name_new_folder)

    basic_functions.create_folder(
        path_result_sub_folder)

    path = os.path.join(
        path_result_sub_folder,
        "{}__{}{}".format(file_name_peak_day, sim_yr, ".npy"))

    np.save(path, model_results)

def write_supply_results(
        sim_yr,
        name_new_folder,
        path_result,
        model_results,
        file_name
    ):
    """Write model results to numpy file as follows:

        name of file: name_year
        array in file:  np.array(region, fueltype, timesteps)

    Arguments
    ---------
    sim_yr : int
        Simulation year
    name_new_folder : str
        Name of folder to create
    path_result : str
        Paths
    model_results : array
        Results to store to txt
    file_name : str
        File name
    """
    path_result_sub_folder = os.path.join(
        path_result, name_new_folder)

    basic_functions.create_folder(
        path_result_sub_folder)

    path_file = os.path.join(
        path_result_sub_folder,
        "{}__{}__{}".format(file_name, sim_yr, ".npy"))

    np.save(path_file, model_results)

def write_full_results(
        sim_yr,
        path_result,
        full_results,
        filename
    ):
    """Write out enduse specific results for every hour and store to
    `.npy` file

    Arguments
    -----------
    sim_yr : int
        Simulation year
    path_result : str
        Path
    full_results : dict
        Modelling results per submodel, enduse, region, fueltype, 8760h
    filename : str
        File name
    """
    statistics_to_print = ["{}\t \t \t \t{}".format(
        "Enduse", "total_annual_GWh")]

    # Create folder for model simulation year
    basic_functions.create_folder(path_result)

    basic_functions.create_folder(
        path_result, "full_results")

    for sector_nr in full_results:

        for enduse, fuel in full_results[sector_nr].items():

            path_file = os.path.join(
                os.path.join(path_result, "full_results"),
                "{}__{}__{}__{}__{}".format(
                    filename,
                    enduse,
                    sim_yr,
                    sector_nr,
                    ".npy"))

            np.save(path_file, fuel)

            statistics_to_print.append("{}\t\t\t\t{}".format(
                enduse, np.sum(fuel)))

    # Create statistic files with sum of all end uses
    path_file = os.path.join(
        os.path.join(path_result, "full_results"),
        "{}__{}__{}".format("statistics_end_uses", sim_yr, ".txt"))

    write_list_to_txt(
        path_file,
        statistics_to_print)

def write_residential_tot_demands(
        sim_yr,
        path_result,
        tot_fuel_y_enduse_specific_yh,
        filename
    ):
    basic_functions.create_folder(path_result)
    basic_functions.create_folder(path_result, "residential_results")

    path_file = os.path.join(
        os.path.join(path_result, "residential_results"),
        "{}__{}__{}".format(
            filename,
            sim_yr,
            ".npy"))

    np.save(path_file, tot_fuel_y_enduse_specific_yh)

def write_space_and_water_heating(
        sim_yr,
        path_result,
        tot_fuel_y_enduse_specific_yh,
        filename
    ):
    """Write out enduse specific results for every hour and store to
    `.npy` file

    Arguments
    -----------
    sim_yr : int
        Simulation year
    path_result : str
        Path
    tot_fuel_y_enduse_specific_yh : dict
        Modelling results
    filename : str
        File name
    """
    statistics_to_print = []
    statistics_to_print.append("{}\t \t \t \t{}".format(
        "Enduse", "total_annual_GWh"))

    # Create folder for model simulation year
    basic_functions.create_folder(path_result)

    basic_functions.create_folder(
        path_result, "enduse_specific_results")

    for _, fuel in tot_fuel_y_enduse_specific_yh.items():
        shape = fuel.shape
        fuel_aggregated = np.zeros(shape)
        break

    for enduse, fuel in tot_fuel_y_enduse_specific_yh.items():
        logging.info("   ... Enduse specific writing to file: %s  Total demand: %s ", enduse, np.sum(fuel))

        if enduse in ['rs_space_heating', 'ss_space_heating', 'is_space_heating', 'rs_water_heating', 'ss_water_heating']:
            fuel_aggregated += fuel
    
    path_file = os.path.join(
        os.path.join(path_result, "enduse_specific_results"),
        "{}__{}__{}{}".format(
            filename,
            "all_heating",
            sim_yr,
            ".npy"))

    # Find peak day of electricity across all heating end uses
    lookups = lookup_tables.basic_lookups()
    fueltype_int = lookups['fueltypes']['electricity']

    peak_day_electricity, _ = enduse_func.get_peak_day_single_fueltype(fuel_aggregated[fueltype_int])
    selected_hours = date_prop.convert_yearday_to_8760h_selection(peak_day_electricity)
    selected_demand = fuel_aggregated[:, selected_hours]
    np.save(path_file, selected_demand)

    statistics_to_print.append("{}\t\t\t\t{}".format(
        "all_heating", np.sum(fuel_aggregated)))

    # Create statistic files with sum of all end uses
    path_file = os.path.join(
        os.path.join(path_result, "enduse_specific_results"),
        "{}__{}__{}".format(
            "statistics_end_uses",
            sim_yr,
            ".txt"))

    write_list_to_txt(
        path_file,
        statistics_to_print)

def write_enduse_specific(
        sim_yr,
        path_result,
        tot_fuel_y_enduse_specific_yh,
        filename
    ):
    """Write out enduse specific results for every hour and store to
    `.npy` file

    Arguments
    -----------
    sim_yr : int
        Simulation year
    path_result : str
        Path
    tot_fuel_y_enduse_specific_yh : dict
        Modelling results
    filename : str
        File name
    """
    statistics_to_print = []
    statistics_to_print.append("{}\t \t \t \t{}".format(
        "Enduse", "total_annual_GWh"))

    # Create folder for model simulation year
    basic_functions.create_folder(path_result)

    basic_functions.create_folder(
        path_result, "enduse_specific_results")

    for enduse, fuel in tot_fuel_y_enduse_specific_yh.items():
        logging.info("   ... Enduse specific writing to file: %s  Total demand: %s ", enduse, np.sum(fuel))

        path_file = os.path.join(
            os.path.join(path_result, "enduse_specific_results"),
            "{}__{}__{}__{}".format(
                filename,
                enduse,
                sim_yr,
                ".npy"))

        np.save(path_file, fuel)

        statistics_to_print.append("{}\t\t\t\t{}".format(
            enduse, np.sum(fuel)))

    # Create statistic files with sum of all end uses
    path_file = os.path.join(
        os.path.join(path_result, "enduse_specific_results"),
        "{}__{}__{}".format(
            "statistics_end_uses",
            sim_yr,
            ".txt"))

    write_list_to_txt(
        path_file,
        statistics_to_print)

def write_max_results(sim_yr, path_result, result_foldername, model_results, filename):
    """Store yearly model resuls to numpy array '.npy'

    Arguments
    ---------
    sim_yr : int
        Simulation year
    path_result : str
        Result path
    result_foldername : str
        Folder name
    model_results : np.array
        Model results
    filename : str
        File name

    """
    # Create folder and subolder
    basic_functions.create_folder(path_result)
    basic_functions.create_folder(path_result, result_foldername)

    # Write to txt
    path_file = os.path.join(
        os.path.join(path_result, result_foldername),
        "{}__{}__{}".format(filename, sim_yr, ".npy"))

    np.save(path_file, model_results)

    return

def create_txt_shapes(
        end_use,
        path_txt_shapes,
        shape_peak_dh,
        shape_non_peak_y_dh,
        shape_non_peak_yd
    ):
    """Function collecting functions to write out arrays
    to txt files
    """
    write_array_to_txt(
        os.path.join(
            path_txt_shapes,
            str(end_use) + str("__") + str('shape_peak_dh') + str('.txt')),
        shape_peak_dh)

    write_array_to_txt(
        os.path.join(
            path_txt_shapes,
            str(end_use) + str("__") + str('shape_non_peak_y_dh') + str('.txt')),
        shape_non_peak_y_dh)

    write_array_to_txt(
        os.path.join(
            path_txt_shapes,
            str(end_use) + str("__") + str('shape_non_peak_yd') + str('.txt')),
        shape_non_peak_yd)

    return

def create_csv_file(path, rows):
    """
    #filewriter.writerow(['Spam', 'Lovely Spam', 'Wonderful Spam'])
    """
    with open(path, 'w', newline='') as csvfile:

        filewriter = csv.writer(
            csvfile,
            delimiter=',',
            quotechar='|')

        for row in rows:
            filewriter.writerow(row)

def write_result_txt(path, regions, values):
    """Store values of spatial plot in csv file
    """
    myData = [["region", "value"]]

    for region, value in zip(regions, values):
        myData.append([region, value])

    myFile = open(path, 'w', newline='')
    with myFile:
        writer = csv.writer(myFile)
        writer.writerows(myData)




import os
import numpy as np
import geopandas as gpd
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.patches import Circle
from matplotlib.colors import LinearSegmentedColormap
import argparse
import cartopy.crs as ccrs
import cartopy.io.shapereader as shpreader
from collections import defaultdict

from energy_demand.plotting import fig_p2_weather_val, result_mapping
from energy_demand.basic import basic_functions
from energy_demand.technologies import tech_related
from energy_demand.read_write import write_data
from energy_demand.basic import conversions
from energy_demand.plotting import basic_plot_functions

def plot_4_cross_map(
        cmap_rgb_colors,
        reclassified,
        result_path,
        path_shapefile_input,
        threshold=None,
        seperate_legend=False
    ):
    """Plot classifed 4 cross map
    """
    # --------------
    # Use Cartopy to plot geometreis with reclassified faceolor
    # --------------
    plt.figure(figsize=basic_plot_functions.cm2inch(10, 10)) #, dpi=150)
    proj = ccrs.OSGB() #'epsg:27700'
    ax = plt.axes(projection=proj)
    ax.outline_patch.set_visible(False)

    # set up a dict to hold geometries keyed by our key
    geoms_by_key = defaultdict(list)

    # for each records, pick out our key's value from the record
    # and store the geometry in the relevant list under geoms_by_key
    for record in shpreader.Reader(path_shapefile_input).records():
        region_name = record.attributes['name']
        geoms_by_key[region_name].append(record.geometry)

    # now we have all the geometries in lists for each value of our key
    # add them to the axis, using the relevant color as facecolor
    for key, geoms in geoms_by_key.items():
        region_reclassified_value = reclassified.loc[key]['reclassified']
        facecolor = cmap_rgb_colors[region_reclassified_value]
        ax.add_geometries(geoms, crs=proj, edgecolor='black', facecolor=facecolor, linewidth=0.1)

    # --------------
    # Create Legend
    # --------------
    legend_handles = [
        mpatches.Patch(color=cmap_rgb_colors[0], label=str("+- threshold {}".format(threshold))),
        mpatches.Patch(color=cmap_rgb_colors[1], label=str("a")),
        mpatches.Patch(color=cmap_rgb_colors[2], label=str("b")),
        mpatches.Patch(color=cmap_rgb_colors[3], label=str("c")),
        mpatches.Patch(color=cmap_rgb_colors[4], label=str("d"))]

    legend = plt.legend(
        handles=legend_handles,
        #title="test",
        prop={'size': 8},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.05),
        frameon=False)

    if seperate_legend:
        basic_plot_functions.export_legend(
            legend,
            os.path.join(result_path, "{}__legend.pdf".format(result_path)))
        legend.remove()

    # Remove coordinates from figure
    ax.set_yticklabels([])
    ax.set_xticklabels([])
    legend.get_title().set_fontsize(8)

    # --------
    # Labeling
    # --------
    plt.tight_layout()
    plt.savefig(os.path.join(result_path))
    plt.close()

def plot_4_cross_map_OLD(
        cmap_rgb_colors,
        reclassified,
        result_path,
        path_shapefile_input,
        threshold=None,
        seperate_legend=False
    ):
    """Plot classifed 4 cross map
    """
    # Load uk shapefile
    uk_shapefile = gpd.read_file(path_shapefile_input)

    # Merge stats to geopanda
    shp_gdp_merged = uk_shapefile.merge(reclassified, on='name')

    # Assign projection
    crs = {'init': 'epsg:27700'} #27700: OSGB_1936_British_National_Grid
    uk_gdf = gpd.GeoDataFrame(shp_gdp_merged, crs=crs)
    ax = uk_gdf.plot()

    uk_gdf['facecolor'] = 'white'

    for region in uk_gdf.index:
        reclassified_value = uk_gdf.loc[region]['reclassified']
        uk_gdf.loc[region, 'facecolor'] = cmap_rgb_colors[reclassified_value]

    # plot with face color attribute
    uk_gdf.plot(ax=ax, facecolor=uk_gdf['facecolor'], edgecolor='black', linewidth=0.1)

    legend_handles = [
        mpatches.Patch(color=cmap_rgb_colors[0], label=str("+- thr {}".format(threshold))),
        mpatches.Patch(color=cmap_rgb_colors[1], label=str("a")),
        mpatches.Patch(color=cmap_rgb_colors[2], label=str("b")),
        mpatches.Patch(color=cmap_rgb_colors[3], label=str("c")),
        mpatches.Patch(color=cmap_rgb_colors[4], label=str("d"))]

    legend = plt.legend(
        handles=legend_handles,
        #title="test",
        prop={'size': 8},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.05),
        frameon=False)

    if seperate_legend:
        basic_plot_functions.export_legend(
            legend,
            os.path.join(result_path, "{}__legend.pdf".format(result_path)))
        legend.remove()

    # Remove coordinates from figure
    ax.set_yticklabels([])
    ax.set_xticklabels([])
    legend.get_title().set_fontsize(8)

    # --------
    # Labeling
    # --------
    plt.tight_layout()
    plt.savefig(os.path.join(result_path))
    plt.close()

def total_annual_demand(
        df_data_input,
        path_shapefile_input,
        regions,
        pop_data,
        simulation_yr_to_plot,
        result_path,
        fig_name,
        field_to_plot,
        unit='GW',
        seperate_legend=True,
        bins=False
    ):
    """
    """
    if unit == 'GW':
        conversion_factor = 1
    elif unit == 'kW':
        conversion_factor = conversions.gwh_to_kwh(gwh=1) #GW to KW
    elif unit == 'percentage':
        conversion_factor = 1
    else:
        raise Exception("Not defined unit")

    df_data_input = df_data_input * conversion_factor

    # Load uk shapefile
    uk_shapefile = gpd.read_file(path_shapefile_input)

    # Population of simulation year
    pop_sim_yr = pop_data[simulation_yr_to_plot]

    regions = list(df_data_input.columns)
    nr_of_regions = df_data_input.shape[1]
    nr_of_realisations = df_data_input.shape[0]

    # Mean over all realisations
    mean = df_data_input.mean(axis=0)

    # Mean normalized with population
    mean_norm_pop = df_data_input.mean(axis=0) / pop_sim_yr

    # Standard deviation over all realisations
    std_dev = df_data_input.std(axis=0)

    max_entry = df_data_input.max(axis=0) #maximum entry for every hour
    min_entry = df_data_input.min(axis=0) #maximum entry for every hour

    #print("---- Calculate average per person")
    tot_person = sum(pop_sim_yr)
    #print(df_data_input.iloc[0])
    tot_demand = sum(df_data_input.iloc[0])
    ##print("TOT PERSON: " + str(tot_person))
    #print("TOT PERSON: " + str(tot_demand))
    #print('AVERAGE KW per Person " '+ str(tot_demand / tot_person))

    #print(df_data_input)
    regional_statistics_columns = [
        'name',
        'mean',
        'mean_norm_pop',
        #'mean_norm_pop_std_dev',
        'std_dev']#
        #'diff_av_max',
        #'mean_pp',
        #'diff_av_max_pp',
        #'std_dev_average_every_h',
        #'std_dev_peak_h_norm_pop']

    df_stats = pd.DataFrame(columns=regional_statistics_columns)

    for region_name in regions:

        line_entry = [[
            str(region_name),
            mean[region_name],
            mean_norm_pop[region_name],
            #mean_norm_pop_std_dev[region_name],
            std_dev[region_name]
            #diff_av_max,
            #mean_peak_h_pp,
            #diff_av_max_pp,
            #std_dev_average_every_h,
            #std_dev_peak_h_norm_pop
            ]]

        line_df = pd.DataFrame(
            line_entry,
            columns=regional_statistics_columns)

        df_stats = df_stats.append(line_df)

    # ---------------
    # Create spatial maps
    # http://darribas.org/gds15/content/labs/lab_03.html
    # http://nbviewer.jupyter.org/gist/jorisvandenbossche/57d392c085901eb4981054402b37b6b1
    # ---------------
    # Merge stats to geopanda
    shp_gdp_merged = uk_shapefile.merge(df_stats, on='name')

    # Assign projection
    crs = {'init': 'epsg:27700'} #27700: OSGB_1936_British_National_Grid
    uk_gdf = gpd.GeoDataFrame(shp_gdp_merged, crs=crs)
    ax = uk_gdf.plot()

    # Assign bin colors according to defined cmap and whether
    # plot with min_max values or only min/max values
    #bin_values = [0, 0.0025, 0.005, 0.0075, 0.01]
    nr_of_intervals = 6

    if bins:
        bin_values = bins
    else:
        bin_values = result_mapping.get_reasonable_bin_values_II(
            data_to_plot=list(uk_gdf[field_to_plot]),
            nr_of_intervals=nr_of_intervals)
    print("field_to_plot: {} BINS: {}".format(field_to_plot, bin_values))

    uk_gdf, cmap_rgb_colors, color_zero, min_value, max_value = fig_p2_weather_val.user_defined_bin_classification(
        uk_gdf,
        field_to_plot,
        bin_values=bin_values)

    # plot with face color attribute
    uk_gdf.plot(ax=ax, facecolor=uk_gdf['bin_color'], edgecolor='black', linewidth=0.1)

    # TODO IMRPVE: MAKE CORRECT ONE FOR NEW PROCESSING
    legend_handles = result_mapping.get_legend_handles(
        bin_values[1:-1],
        cmap_rgb_colors,
        color_zero,
        min_value,
        max_value)

    legend = plt.legend(
        handles=legend_handles,
        title="Unit: {} field: {}".format(unit, field_to_plot),
        prop={'size': 8},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.05),
        frameon=False)

    if seperate_legend:
        basic_plot_functions.export_legend(
            legend,
            os.path.join(result_path, "{}__legend.pdf".format(fig_name)))
        legend.remove()

    # Remove coordinates from figure
    ax.set_yticklabels([])
    ax.set_xticklabels([])

    legend.get_title().set_fontsize(8)

    # PLot bins on plot
    '''plt.text(
        0,
        -20,
        bin_values[:-1],
        fontsize=8)'''

    # --------
    # Labeling
    # --------
    #plt.title("Peak demand over time")
    plt.tight_layout()
    #plt.show()
    plt.savefig(os.path.join(result_path, fig_name))
    plt.close()



"""Plotting model results and storing as PDF to result folder
"""
import os
import logging

from energy_demand.technologies import tech_related

from energy_demand.plotting import plotting_styles
from energy_demand.plotting import fig_lad_related
from energy_demand.plotting import fig_one_fueltype_multiple_regions_peak_h
from energy_demand.plotting import fig_fuels_enduses_y
from energy_demand.plotting import fig_stacked_enduse
from energy_demand.plotting import fig_cross_graphs
from energy_demand.plotting import fig_stacked_enduse_sectors
from energy_demand.plotting import fig_lf
from energy_demand.plotting import fig_fuels_enduses_week
from energy_demand.plotting import fig_load_profile_dh_multiple
from energy_demand.plotting import fig_fuels_peak_h
from energy_demand.plotting import fig_weather_variability_priod

#matplotlib.use('Agg') # Used to make it work in linux179

def run_all_plot_functions(
        results_container,
        reg_nrs,
        regions,
        lookups,
        result_paths,
        assumptions,
        enduses,
        plot_crit,
        base_yr,
        comparison_year
    ):
    """Summary function to plot all results

    comparison_year : int
        Year to generate comparison plots
    """
    if plot_crit['plot_lad_cross_graphs']:

        try:
            # Plot cross graph where very region is a dot
            fig_cross_graphs.plot_cross_graphs(
                base_yr=base_yr,
                comparison_year=comparison_year,
                regions=regions,
                ed_year_fueltype_regs_yh=results_container['ed_fueltype_regs_yh'],
                reg_load_factor_y=results_container['reg_load_factor_y'],
                fueltype_int=lookups['fueltypes']['electricity'],
                fueltype_str='electricity',
                fig_name=os.path.join(
                    result_paths['data_results_PDF'], "comparions_LAD_cross_graph_electricity_by_cy.pdf"),
                label_points=False,
                plotshow=False)

            fig_cross_graphs.plot_cross_graphs(
                base_yr=base_yr,
                comparison_year=comparison_year,
                regions=regions,
                ed_year_fueltype_regs_yh=results_container['ed_fueltype_regs_yh'],
                reg_load_factor_y=results_container['reg_load_factor_y'],
                fueltype_int=lookups['fueltypes']['gas'],
                fueltype_str='gas',
                fig_name=os.path.join(
                    result_paths['data_results_PDF'], "comparions_LAD_cross_graph_gas_by_cy.pdf"),
                label_points=False,
                plotshow=False)
        except KeyError:
            logging.info("Check if correct comparison year is provided, i.e. really data exists for this year")

    # ----------
    # Plot LAD differences for first and last year
    # ----------
    try:
        fig_lad_related.plot_lad_comparison(
            base_yr=2015,
            comparison_year=2050,
            regions=regions,
            ed_year_fueltype_regs_yh=results_container['ed_fueltype_regs_yh'],
            fueltype_int=lookups['fueltypes']['electricity'],
            fueltype_str='electricity',
            fig_name=os.path.join(
                result_paths['data_results_PDF'], "comparions_LAD_modelled_electricity_by_cy.pdf"),
            label_points=False,
            plotshow=False)
        print("... plotted by-cy LAD energy demand compariosn")

        # Plot peak h for every hour
        fig_lad_related.lad_comparison_peak(
            base_yr=2015,
            comparison_year=2050,
            regions=regions,
            ed_year_fueltype_regs_yh=results_container['ed_fueltype_regs_yh'],
            fueltype_int=lookups['fueltypes']['electricity'],
            fueltype_str='electricity',
            fig_name=os.path.join(
                result_paths['data_results_PDF'], "comparions_LAD_modelled_electricity_peakh_by_cy.pdf"),
            label_points=False,
            plotshow=False)
        print("... plotted by-cy LAD energy demand compariosn")
    except:
        pass

    # ----------------
    # Plot demand for every region over time
    # -------------------
    if plot_crit['plot_line_for_every_region_of_peak_demand']:
        logging.info("... plot fuel per fueltype for every region over annual teimsteps")
        fig_one_fueltype_multiple_regions_peak_h.plt_regions_peak_h(
            results_container['ed_fueltype_regs_yh'],
            lookups,
            regions,
            os.path.join(
                result_paths['data_results_PDF'],
                'peak_h_total_electricity.pdf'),
            fueltype_str_to_plot="electricity")

    if plot_crit['plot_fuels_enduses_y']:

        #... Plot total fuel (y) per fueltype as line chart"
        fig_fuels_enduses_y.run(
            results_container['ed_fueltype_regs_yh'],
            lookups,
            os.path.join(
                result_paths['data_results_PDF'],
                'y_fueltypes_all_enduses.pdf'))

    # ------------
    # Plot stacked annual enduses
    # ------------
    if plot_crit['plot_stacked_enduses']:

        rs_enduses_sorted = [
            'rs_space_heating',
            'rs_water_heating',
            'rs_lighting',
            'rs_cold',
            'rs_wet',
            'rs_consumer_electronics',
            'rs_home_computing',
            'rs_cooking']
    
        ss_enduses_sorted = [
            'ss_space_heating',
            'ss_water_heating',
            'ss_lighting',
            'ss_catering',
            'ss_small_power',
            'ss_fans',
            'ss_cooling_humidification',
            'ss_ICT_equipment',
            'ss_other_gas',
            'ss_other_electricity',
            'ss_cooled_storage']
        
        is_enduses_sorted = [
            'is_space_heating',
            'is_lighting',
            'is_refrigeration',
            'is_motors',
            'is_compressed_air',
            'is_high_temp_process',
            'is_low_temp_process',
            'is_other',
            'is_drying_separation']

        rs_color_list = plotting_styles.rs_color_list_selection()
        ss_color_list = plotting_styles.ss_color_list_selection()
        is_color_list = plotting_styles.is_color_list_selection()

        # Residential
        fig_stacked_enduse.run(
            assumptions['sim_yrs'],
            results_container['results_enduse_every_year'],
            rs_enduses_sorted,
            rs_color_list,
            os.path.join(
                result_paths['data_results_PDF'], "stacked_rs_country.pdf"),
            plot_legend=True)

        # Service
        fig_stacked_enduse.run(
            assumptions['sim_yrs'],
            results_container['results_enduse_every_year'],
            ss_enduses_sorted,
            ss_color_list,
            os.path.join(
                result_paths['data_results_PDF'], "stacked_ss_country.pdf"),
            plot_legend=True)

        # Industry
        fig_stacked_enduse.run(
            assumptions['sim_yrs'],
            results_container['results_enduse_every_year'],
            is_enduses_sorted,
            is_color_list,
            os.path.join(
                result_paths['data_results_PDF'], "stacked_is_country_.pdf"),
            plot_legend=True)

    # ------------------------------
    # Plot annual demand for enduses for all submodels
    # ------------------------------
    if plot_crit['plot_y_all_enduses']:

        fig_stacked_enduse_sectors.run(
            lookups,
            assumptions['sim_yrs'],
            results_container['results_enduse_every_year'],
            enduses['residential'],
            enduses['service'],
            enduses['industry'],
            os.path.join(result_paths['data_results_PDF'],
            "stacked_all_enduses_country.pdf"))

    # --------------
    # Fuel per fueltype for whole country over annual timesteps
    # ----------------
    if plot_crit['plot_fuels_enduses_y']:
        logging.info("... plot fuel per fueltype for whole country over annual timesteps")
        #... Plot total fuel (y) per fueltype as line chart"
        fig_fuels_enduses_y.run(
            results_container['ed_fueltype_regs_yh'],
            lookups,
            os.path.join(
                result_paths['data_results_PDF'],
                'y_fueltypes_all_enduses.pdf'))

    # ----------
    # Plot seasonal typical load profiles
    # Averaged load profile per daytpe for a region
    # ----------

    # ------------------------------------
    # Load factors per fueltype and region
    # ------------------------------------
    if plot_crit['plot_lf'] :
        for fueltype_str, fueltype_int in lookups['fueltypes'].items():
            '''fig_lf.plot_seasonal_lf(
                fueltype_int,
                fueltype_str,
                results_container['load_factor_seasons'],
                reg_nrs,
                os.path.join(
                    result_paths['data_results_PDF'],
                    'lf_seasonal_{}.pdf'.format(fueltype_str)))'''

            '''fig_lf.plot_lf_y(
                fueltype_int,
                fueltype_str,
                results_container['reg_load_factor_yd'],
                reg_nrs,
                os.path.join(
                    result_paths['data_results_PDF'], 'lf_yd_{}.pdf'.format(fueltype_str)))'''

            # reg_load_factor_yd = max daily value / average annual daily value
            fig_lf.plot_lf_y(
                fueltype_int,
                fueltype_str,
                results_container['reg_load_factor_y'],
                reg_nrs,
                os.path.join(
                    result_paths['data_results_PDF'],
                    'lf_y_{}.pdf'.format(fueltype_str)))

    # --------------
    # Fuel week of base year
    # ----------------
    if plot_crit['plot_week_h']:
        fig_fuels_enduses_week.run(
            results_resid=results_container['ed_fueltype_regs_yh'],
            lookups=lookups,
            hours_to_plot=range(7*24),
            year_to_plot=2015,
            fig_name=os.path.join(result_paths['data_results_PDF'], "tot_all_enduse03.pdf"))

    # ------------------------------------
    # Plot averaged per season and fueltype
    # ------------------------------------
    if plot_crit['plot_averaged_season_fueltype']:
        for year in results_container['av_season_daytype_cy'].keys():
            for fueltype_int in results_container['av_season_daytype_cy'][year].keys():

                fueltype_str = tech_related.get_fueltype_str(
                    lookups['fueltypes'], fueltype_int)

                fig_load_profile_dh_multiple.run(
                    path_fig_folder=result_paths['data_results_PDF'],
                    path_plot_fig=os.path.join(
                        result_paths['data_results_PDF'],
                        'season_daytypes_by_cy_comparison__{}__{}.pdf'.format(year, fueltype_str)),
                    calc_av_lp_modelled=results_container['av_season_daytype_cy'][year][fueltype_int],  # current year
                    calc_av_lp_real=results_container['av_season_daytype_cy'][base_yr][fueltype_int], # base year
                    calc_lp_modelled=results_container['season_daytype_cy'][year][fueltype_int],        # current year
                    calc_lp_real=results_container['season_daytype_cy'][base_yr][fueltype_int],       # base year
                    plot_peak=True,
                    plot_all_entries=False,
                    plot_max_min_polygon=True,
                    plotshow=False,
                    plot_radar=plot_crit['plot_radar_seasonal'],
                    max_y_to_plot=120,
                    fueltype_str=fueltype_str,
                    year=year)

    # ---------------------------------
    # Plot hourly peak loads over time for different fueltypes
    # --------------------------------
    if plot_crit['plot_h_peak_fueltypes']:

        fig_fuels_peak_h.run(
            results_container['ed_fueltype_regs_yh'],
            lookups,
            os.path.join(
                result_paths['data_results_PDF'],
                'fuel_fueltypes_peak_h.pdf'))

    print("finisthed plotting")
    return



import os
import numpy as np
import pandas as pd

def create_annual_weather_file(
        path,
        path_out_stations,
        year,
        weather_scenario_name
    ):
    """Create annual weather file

    write csv file
    """
    # Read files
    path_stations = os.path.join(path, "{}_stations.csv".format(year))
    path_t_min = os.path.join(path, "{}_t_min.npy".format(year))
    path_t_max = os.path.join(path, "{}_t_max.npy".format(year))
    df_stations = pd.read_csv(path_stations)
    stations = df_stations['station_id'].values.tolist()
    t_min = np.load(path_t_min)
    t_max = np.load(path_t_max)

    out_csv = []
    for station_cnt, station_id in enumerate(stations):
        t_min_station = t_min[station_cnt]
        t_max_station = t_max[station_cnt]
        for yearday in range(365):
            out_csv.append([year, station_id, weather_scenario_name, yearday, t_min_station[yearday], t_max_station[yearday]])

    out_csv_array = np.array(out_csv)
    return out_csv_array

def collect_multi_year_weather(path, path_out, year):
    
    weather_scenario_name = os.path.join(path_out, "weather_temps_{}".format(year))
    path_out_csv = os.path.join(weather_scenario_name + ".csv")

    out_csv_array = create_annual_weather_file(
        path,
        path_out,
        year=2015,
        weather_scenario_name=weather_scenario_name)
        
    columns = ['timestep', 'station_id', 'stiching_name', 'yearday', 't_min', 't_max',]
    df = pd.DataFrame(out_csv_array, columns=columns)
    df.to_csv(path_out_csv, index=False)


path = "X:/nismod/data/energy_demand/H-Met_office_weather_data/_complete_meteo_data_all_yrs_cleaned_min_max"
path_out = "X:/nismod/data/energy_demand/H-Met_office_weather_data/_complete_meteo_data_all_yrs_cleaned_min_max"

collect_multi_year_weather(
    path,
    path_out,
    year=2015)


"""
"""
import os
import operator
import logging
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
import pandas as pd

from energy_demand.technologies import tech_related
from energy_demand.read_write import data_loader
from energy_demand.plotting import basic_plot_functions
from energy_demand.plotting import plotting_styles
from energy_demand.validation import lad_validation

def run(
        simulation_yr_to_plot,
        demand_year_non_regional,
        demand_year_regional,
        fueltypes,
        fig_path,
        path_temporal_elec_validation,
        path_temporal_gas_validation,
        regions,
        plot_crit
    ):
    """Validate spatial and temporal energy demands

    Info
    -----
    Because the floor area is only availabe for LADs from 2001,
    the LADs are converted to 2015 LADs.
    """
    logging.info("... temporal validation")

    ######################################
    # Data preparation
    ######################################
    subnational_elec = data_loader.read_lad_demands(path_temporal_elec_validation) # Read only domestic heating
    subnational_gas = data_loader.read_lad_demands(path_temporal_gas_validation)   # Read only domestic heating

    # ----------------------------------------
    # Remap demands between 2011 and 2015 LADs
    # ----------------------------------------
    subnational_elec = lad_validation.map_LAD_2011_2015(subnational_elec)
    subnational_gas = lad_validation.map_LAD_2011_2015(subnational_gas)

    # Create fueltype secific dict electricity
    fuel_elec_regs_yh_non_regional = {}
    for region_array_nr, region in enumerate(regions):
        gwh_modelled_tot = demand_year_non_regional[fueltypes['electricity']][region_array_nr]
        fuel_elec_regs_yh_non_regional[region] = gwh_modelled_tot

    # Create fueltype secific dict for gas
    fuel_gas_regs_yh = {}
    for region_array_nr, region in enumerate(regions):
        gwh_modelled_tot = np.sum(demand_year_non_regional[fueltypes['gas']][region_array_nr])
        fuel_gas_regs_yh[region] = gwh_modelled_tot

    fuel_elec_regs_yh_non_regional = lad_validation.map_LAD_2011_2015(fuel_elec_regs_yh_non_regional)
    fuel_gas_regs_yh = lad_validation.map_LAD_2011_2015(fuel_gas_regs_yh)

    ######################################
    # Plotting
    ######################################
    # Electrcity
    plot_spatial_validation(
        simulation_yr_to_plot,
        fuel_elec_regs_yh_non_regional,
        demand_year_regional,
        subnational_elec,
        regions,
        'electricity',
        fig_path=os.path.join(fig_path, "spatial_validation_electricity.pdf"),
        label_points=False,
        plotshow=plot_crit)

    # Gas
    plot_spatial_validation(
        simulation_yr_to_plot,
        fuel_gas_regs_yh,
        demand_year_regional,
        subnational_gas,
        regions,
        'gas',
        fig_path=os.path.join(fig_path, "spatial_validation_gas.pdf"),
        label_points=False,
        plotshow=plot_crit)

def plot_spatial_validation(
        simulation_yr_to_plot,
        non_regional_modelled,
        regional_modelled,
        subnational_real,
        regions,
        fueltype_str,
        fig_path,
        label_points=False,
        plotshow=False
    ):
    result_dict = {}
    result_dict['real_demand'] = {}
    result_dict['modelled_demand'] = {}
    result_dict['modelled_demands_regional'] = {}

    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    # -------------------------------------------
    # Match ECUK sub-regional demand with geocode
    # -------------------------------------------
    for region_nr, region in enumerate(regions):
        try:
            # --Sub Regional Electricity demand (as GWh)
            real = subnational_real[region]
            modelled = non_regional_modelled[region]
            result_dict['real_demand'][region] = real
            result_dict['modelled_demand'][region] = modelled

        except KeyError:
            logging.debug("Sub-national spatial validation: No fuel for region %s", region)

        # Do this for every weather station data
        for weather_station in regional_modelled:
            try:
                _reg_demand = regional_modelled[weather_station][simulation_yr_to_plot][fueltype_int][region_nr]
                result_dict['modelled_demands_regional'][region].append(_reg_demand)
            except KeyError:
                _reg_demand = regional_modelled[weather_station][simulation_yr_to_plot][fueltype_int][region_nr]
                result_dict['modelled_demands_regional'][region] = [_reg_demand]

    # --------------------
    # Calculate statistics
    # --------------------
    diff_real_modelled_p = []
    diff_real_modelled_abs = []

    for region in regions:
        try:
            real = result_dict['real_demand'][region]
            modelled = result_dict['modelled_demand'][region]
            diff_real_modelled_p.append(abs(100 - ((100 / real) * modelled)))
            diff_real_modelled_abs.append(real - modelled)
        except KeyError:
            pass

    # Calculate the average deviation between reald and modelled
    av_deviation_real_modelled = np.average(diff_real_modelled_p)
    median_absolute_deviation = np.median(diff_real_modelled_p)    # median deviation

    # Calculate standard deviation
    std_dev_p = np.std(diff_real_modelled_p)        # Given as percent
    std_dev_abs = np.std(diff_real_modelled_abs)    # Given as energy unit

    # -----------------
    # Sort results according to size
    # -----------------
    sorted_dict_real = sorted(
        result_dict['real_demand'].items(),
        key=operator.itemgetter(1))

    # -------------------------------------
    # Plot
    # -------------------------------------
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(9, 8))
    ax = fig.add_subplot(1, 1, 1)
    x_values = np.arange(0, len(sorted_dict_real), 1)
    y_real_demand = []
    y_modelled_demand = []
    y_modelled_demands_non_regional = []
    labels = []

    for sorted_region in sorted_dict_real:
        geocode_lad = sorted_region[0]
        y_real_demand.append(result_dict['real_demand'][geocode_lad])
        y_modelled_demand.append(result_dict['modelled_demand'][geocode_lad])
        y_modelled_demands_non_regional.append(result_dict['modelled_demands_regional'][geocode_lad])
    
        print(
            "validation %s LAD %s: real: %s modelled: %s  modelled percentage: %s (%sp diff)",
            fueltype_str,
            geocode_lad,
            round(result_dict['real_demand'][geocode_lad], 4),
            round(result_dict['modelled_demand'][geocode_lad], 4),
            round(100 / result_dict['real_demand'][geocode_lad] * result_dict['modelled_demand'][geocode_lad], 4),
            round(100 - (100 / result_dict['real_demand'][geocode_lad] * result_dict['modelled_demand'][geocode_lad]), 4))

        labels.append(geocode_lad)

    # Calculate r_squared
    _slope, _intercept, r_value, _p_value, _std_err = stats.linregress(
        y_real_demand,
        y_modelled_demand)

    # --------
    # Axis
    # --------
    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='off',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='off') # labels along the bottom edge are off

    # ----------------------------------------------
    # Plot
    # ----------------------------------------------
    markersize = 3
    markeredgewidth = 0
    linewidth = 2
    color_real = 'black'
    color_all_stations = 'green'
    color_single_station = 'red'
    plt.plot(
        x_values,
        y_real_demand,
        linestyle='-',
        marker='o',
        alpha=0.6,
        markersize=markersize,
        fillstyle='full',
        markerfacecolor='black',
        markeredgewidth=markeredgewidth,
        color=color_real,
        label='actual demand')

    plt.plot(
        x_values,
        y_modelled_demand,
        marker='o',
        linestyle='-',
        markersize=markersize,
        alpha=0.6,
        markerfacecolor='blue',
        fillstyle='none',
        markeredgewidth=markeredgewidth,
        markeredgecolor='blue',
        color=color_all_stations,
        label='modelled using all stations')

    # Demands calculated only from one weather station
    station_nr = 0
    nr_of_stations = len(y_modelled_demands_non_regional[0])
    station_vals = []
    for region_vals in y_modelled_demands_non_regional:
        station_vals.append(region_vals[station_nr])

    plt.plot(
        x_values,
        station_vals,
        marker='o',
        linestyle='-',
        linewidth=linewidth,
        markersize=markersize,
        alpha=0.6,
        markerfacecolor='green',
        fillstyle='none',
        markeredgewidth=markeredgewidth,
        markeredgecolor='green',
        color=color_single_station,
        label='modelled using only a single stations')

    '''for i in range(nr_of_stations):
        station_data = []
        for reg_nr in range(nr_of_regions):
            station_data.append(y_modelled_demands_non_regional[reg_nr][i])

        plt.plot(
            x_values,
            station_data, #y_modelled_demands_non_regional,
            marker='o',
            linestyle='None',
            markersize=1.6,
            alpha=0.6,
            markerfacecolor='white',
            fillstyle='none',
            markeredgewidth=0.5,
            markeredgecolor='orange',
            color='black',
            label='model')'''

    '''# ------------
    # Collect all values per weather_yr
    list_with_station_vals = []
    for station_i in range(nr_of_stations):
        station_vals = []
        for region_vals in y_modelled_demands_non_regional:
            station_vals.append(region_vals[station_i])
        list_with_station_vals.append(station_vals)

    df = pd.DataFrame(
        list_with_station_vals,
        columns=range(nr_of_regions)) #note not region_rn as ordered

    period_h = range(nr_of_regions)

    quantile_95 = 0.95
    quantile_05 = 0.05

    df_q_95 = df.quantile(quantile_95)
    df_q_05 = df.quantile(quantile_05)

    #Transpose for plotting purposes
    df = df.T
    df_q_95 = df_q_95.T
    df_q_05 = df_q_05.T

    # ---------------
    # Smoothing lines
    # ---------------
    try:
        period_h_smoothed, df_q_95_smoothed = basic_plot_functions.smooth_data(period_h, df_q_95, num=40000)
        period_h_smoothed, df_q_05_smoothed = basic_plot_functions.smooth_data(period_h, df_q_05, num=40000)
    except:
        period_h_smoothed = period_h
        df_q_95_smoothed = df_q_95
        df_q_05_smoothed = df_q_05
    #plt.plot(period_h_smoothed, df_q_05_smoothed, color='black', linestyle='--', linewidth=0.5, label="0.05")
    #plt.plot(period_h_smoothed, df_q_95_smoothed, color='black', linestyle='--', linewidth=0.5, label="0.95")

    # -----------------
    # Uncertainty range
    # -----------------
    plt.fill_between(
        period_h_smoothed, #x
        df_q_95_smoothed,  #y1
        df_q_05_smoothed,  #y2
        alpha=.40,
        facecolor="grey",
        label="uncertainty band")
    # -----------'''


    # Limit
    plt.ylim(ymin=0)

    # -----------
    # Labelling
    # -----------
    if label_points:
        for pos, txt in enumerate(labels):

            ax.text(
                x_values[pos],
                y_modelled_demand[pos],
                txt,
                horizontalalignment="right",
                verticalalignment="top",
                fontsize=1)

    font_additional_info = plotting_styles.font_info(size=4)

    title_info = ('R_2: {}, std_%: {} (GWh {}), av_diff_%: {} median_abs_dev: {}'.format(
        round(r_value, 2),
        round(std_dev_p, 2),
        round(std_dev_abs, 2),
        round(av_deviation_real_modelled, 2),
        round(median_absolute_deviation, 2)))

    plt.title(
        title_info,
        loc='left',
        fontdict=font_additional_info)

    plt.xlabel("UK regions (excluding northern ireland)")
    plt.ylabel("{} [GWh]".format(fueltype_str))

    # --------
    # Legend
    # --------
    plt.legend(
        prop={
            'family': 'arial',
            'size': 8},
        frameon=False)

    # Tight layout
    plt.margins(x=0)
    plt.tight_layout()
    plt.savefig(fig_path)

    if plotshow:
        plt.show()
    else:
        plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for energy_demand.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""
from __future__ import absolute_import, division, print_function

from pytest import fixture


@fixture(scope='function')
def config_file(tmpdir):
    config_file = tmpdir.mkdir("config").join("wrapperconfig.ini")
    contents = """
[PATHS]
path_local_data = tests
path_processed_data = tests/_processed_data
path_result_data = tests/results
path_config_data = tests/config_data
path_new_scenario = tests/results/tmp_model_run_results

[CONFIG]
base_yr = 2015
weather_yr_scenario = 2015
user_defined_simulation_end_yr = 2050
user_defined_weather_by = 2015

[CRITERIA]
mode_constrained = True
virtual_building_stock_criteria = True
write_out_national = False
reg_selection = False
MSOA_crit = False
reg_selection_csv_name = msoa_regions_ed.csv
spatial_calibration = False
cluster_calc = False
write_txt_additional_results = True
validation_criteria = True
plot_crit = False
crit_plot_enduse_lp = False
writeYAML_keynames = True
writeYAML = False
crit_temp_min_max = False

[DATA_PATHS]
local_path_datafolder = tests/energy_demand_minimal
path_strategy_vars = tests/energy_demand_minimal/00_user_defined_variables
#ONS principal projection
path_population_data_for_disaggregation_LAD =  tests/energy_demand_minimal/_raw_data/J-population_disagg_by/uk_pop_principal_2015_2050.csv  
#ONS principal projection
folder_raw_carbon_trust = tests/energy_demand_minimal/_raw_data/G_Carbon_Trust_advanced_metering_trial/
path_population_data_for_disaggregation_MSOA = tests/energy_demand_minimal/_raw_data/J-population_disagg_by/uk_pop_principal_2015_2050_MSOA_lad.csv
folder_path_weater_stations = tests/energy_demand_minimal/_raw_data/A-temperature_data/cleaned_weather_stations.csv
path_floor_area_virtual_stock_by = tests/energy_demand_minimal/_raw_data/K-floor_area/floor_area_LAD_latest.csv
path_assumptions_db = tests/_processed_data/assumptions_from_db
data_processed = tests/_processed_data
lad_shapefile = tests/energy_demand_minimal/_raw_data/C_LAD_geography/same_as_pop_scenario/lad_2016_uk_simplified.shp
path_post_installation_data = tests/_processed_data
weather_data = /raw_data/A-temperature_data/cleaned_weather_stations_data
load_profiles = tests/_processed_data/load_profiles
rs_load_profile_txt = tests/_processed_data/load_profiles/rs_submodel
ss_load_profile_txt = tests/_processed_data/load_profiles/ss_submodel
yaml_parameters = tests/config/yaml_parameters.yml
yaml_parameters_constrained = tests/config/yaml_parameters_constrained.yml
yaml_parameters_keynames_constrained = tests/config/yaml_parameters_keynames_constrained.yml
yaml_parameters_keynames_unconstrained = tests/config/yaml_parameters_keynames_unconstrained.yml
yaml_parameters_scenario = tests/config/yaml_parameters_scenario.yml

[CONFIG_DATA]
path_main = tests/config_data
# Path to all technologies
path_technologies = tests/config_data/05-technologies/technology_definition.csv
# Paths to fuel raw data
rs_fuel_raw = tests/config_data/02-fuel_base_year/rs_fuel.csv
ss_fuel_raw = tests/config_data/02-fuel_base_year/ss_fuel.csv
is_fuel_raw = tests/config_data/02-fuel_base_year/is_fuel.csv
# Load profiles
lp_rs = tests/config_data/03-load_profiles/rs_submodel/HES_lp.csv
# Technologies load shapes
path_hourly_gas_shape_resid = tests/config_data/03-load_profiles/rs_submodel/lp_gas_boiler_dh_SANSOM.csv
lp_elec_hp_dh = tests/config_data/03-load_profiles/rs_submodel/lp_elec_hp_dh_LOVE.csv
lp_all_microCHP_dh = tests/config_data/03-load_profiles/rs_submodel/lp_all_microCHP_dh_SANSOM.csv
path_shape_rs_cooling = tests/config_data/03-load_profiles/rs_submodel/shape_residential_cooling.csv
path_shape_ss_cooling = tests/config_data/03-load_profiles/ss_submodel/shape_service_cooling.csv
lp_elec_storage_heating = tests/config_data/03-load_profiles/rs_submodel/lp_elec_storage_heating_HESReport.csv
lp_elec_secondary_heating = tests/config_data/03-load_profiles/rs_submodel/lp_elec_secondary_heating_HES.csv
# Census data
path_employment_statistics = tests/config_data/04-census_data/LAD_census_data.csv
# Validation datasets
val_subnational_elec = tests/config_data/01-validation_datasets/02_subnational_elec/data_2015_elec.csv
val_subnational_elec_residential = tests/config_data/01-validation_datasets/02_subnational_elec/data_2015_elec_domestic.csv
val_subnational_elec_non_residential = tests/config_data/01-validation_datasets/02_subnational_elec/data_2015_elec_non_domestic.csv
val_subnational_elec_msoa_residential = tests/config_data/01-validation_datasets/02_subnational_elec/MSOA_domestic_electricity_2015_cleaned.csv
val_subnational_elec_msoa_non_residential = tests/config_data/01-validation_datasets/02_subnational_elec/MSOA_non_dom_electricity_2015_cleaned.csv
val_subnational_gas = tests/config_data/01-validation_datasets/03_subnational_gas/data_2015_gas.csv
val_subnational_gas_residential = tests/config_data/01-validation_datasets/03_subnational_gas/data_2015_gas_domestic.csv
val_subnational_gas_non_residential = tests/config_data/01-validation_datasets/03_subnational_gas/data_2015_gas_non_domestic.csv
val_nat_elec_data = tests/config_data/01-validation_datasets/01_national_elec_2015/elec_demand_2015.csv
"""
    config_file.write(contents)
    return str(config_file)



"""Project 2011 MSOA population to the future with help of LAD population scenarios
"""
import os
from collections import defaultdict
from energy_demand.read_write import data_loader
from energy_demand.read_write import write_data

def remap_lads(lad_name):
    """Some LADs of MSOA census data do not match
    the LAD geographies of the LAD population data
    """
    if lad_name == 'E06000048':
        lad_mapped = 'E06000057' #Northumberland
    elif lad_name == 'E08000020':
        lad_mapped = 'E08000037'
    elif lad_name == 'E07000097':
        lad_mapped = 'E07000242'
    elif lad_name == 'E07000100':
        lad_mapped = 'E07000240'
    elif lad_name == 'E07000101':
        lad_mapped = 'E07000243'
    elif lad_name == 'E07000104': #Welwyn Hatfield
        lad_mapped = 'E07000241'
    else:
        lad_mapped = lad_name

    return lad_mapped

mapping_dict = True
population_script = False
gva_script = False

# ------------------------------------------------
# Create mapping dict {"LAD": [MSOAs]}
# ------------------------------------------------
if mapping_dict:
    local_data_path = "C:/Users/cenv0553/ED/data/scenarios"
    path_to_csv_MSOA = os.path.join(local_data_path, "pop_MSOA_orig.csv")
    path_to_csv_LAD = os.path.join(local_data_path, 'uk_pop_high_migration_2015_2050.csv')
    path_out = os.path.join(local_data_path, 'mapping_MSOA_LAD.txt')

    # Read 2011 MSOA population data
    population_2011 = data_loader.load_MOSA_pop(path_to_csv_MSOA)

    # Read 2015 - 2050 population data for LAD
    pop_lads_2015_2015 = data_loader.read_scenario_data(path_to_csv_LAD)

    all_msoas = []
    for lad, msoas in population_2011.items():
        all_msoas += list(msoas.keys())

    # all lads
    all_lads = list(pop_lads_2015_2015[2015].keys())

    dict_mapping = {}

    # ----------------------------------------------------
    # Iterate MSOA and add population
    # ----------------------------------------------------
    for msoa in all_msoas:

        for lad, msoas in population_2011.items():
            msoas_of_lad = list(msoas.keys())
            if msoa in msoas_of_lad:
                lad_match = lad
                break

        # Remap
        lad_match_remap = remap_lads(lad_match)
        try:
            dict_mapping[lad_match_remap].append(msoa)
        except KeyError:
            dict_mapping[lad_match_remap] = [msoa]

    print(len(dict_mapping.keys()))
    text_file = open(path_out, "w")
    text_file.write(str(dict_mapping))
    text_file.close()
    print("Finished dict script")

# ------------------------------------------------
# Population
# ------------------------------------------------
if population_script:
    local_data_path = "C:/Users/cenv0553/ED/data/scenarios"
    path_to_csv_MSOA = os.path.join(local_data_path, "pop_MSOA_orig.csv")
    path_to_csv_LAD = os.path.join(local_data_path, 'uk_pop_high_migration_2015_2050.csv')
    path_out = os.path.join(local_data_path, 'uk_pop_high_migration_2015_2050_MSOA_lad.csv')

    # Read 2011 MSOA population data
    population_2011 = data_loader.load_MOSA_pop(path_to_csv_MSOA)

    # Read 2015 - 2050 population data for LAD
    pop_lads_2015_2015 = data_loader.read_scenario_data(path_to_csv_LAD)

    all_msoas = []
    for lad, msoas in population_2011.items():
        all_msoas += list(msoas.keys())

    msoa_pop_2011_2050 = defaultdict(dict)

    # all lads
    all_lads = list(pop_lads_2015_2015[2015].keys())

    # ----------------------------------------------------
    # Iterate MSOA and add population
    # ----------------------------------------------------
    for msoa in all_msoas:

        for lad, msoas in population_2011.items():
            msoas_of_lad = list(msoas.keys())
            if msoa in msoas_of_lad:
                lad_match = lad
                break

        # Remap
        lad_match_remap = remap_lads(lad_match)

        # Remove lad
        for lad_position, lad in enumerate(all_lads):
            if lad == lad_match_remap:
                all_lads.pop(lad_position)
                break

        # pop_LAD_2011
        lad_pop_by = sum(population_2011[lad_match].values())

        for year in pop_lads_2015_2015.keys():

            # pop msoa 2011
            msoa_pop_2011 = population_2011[lad_match][msoa]

            # pop lad cy
            lad_pop_cy = pop_lads_2015_2015[year][lad_match_remap]

            # Get change in LAd projection and assume that
            # the same change for all MSOA of the corresponding LAD
            change_p = lad_pop_cy / lad_pop_by

            msoa_pop_2011_2050[year][msoa] = change_p * msoa_pop_2011

    # ----------------------------------------------------
    # Iterate all lads and add population for northern ireland and scotland
    # ----------------------------------------------------
    for lad in all_lads:
        for year in pop_lads_2015_2015.keys():
            msoa_pop_2011_2050[year][lad] = pop_lads_2015_2015[year][lad]

    # Write msoa to file
    rows = [("region", "year", "value", "interval")]

    for year, msoas in msoa_pop_2011_2050.items():
        for msoa, pop in msoas.items():
            rows.append([msoa, year, pop, 1])

    write_data.create_csv_file(path_out, rows)

    #np.savetxt(path_out, rows, delimiter=",", header="{}, {}, {}, {}".format("region", "year", "value", "interval"), comments='')
    print("Finished pop script")

# ------------------------------------------------
# GVA Dummy data
# ------------------------------------------------
if gva_script:
    local_data_path = "C:/Users/cenv0553/ED/data/scenarios"
    
    path_to_csv_MSOA = os.path.join(local_data_path, "pop_MSOA_orig.csv")
    path_to_csv_LAD_pop = os.path.join(local_data_path, 'uk_pop_high_migration_2015_2050.csv')
    path_out = os.path.join(local_data_path, 'gva_sven_msoa_lad.csv')

    dummy_gva = 1000

    # Get all MSOA
    path_to_csv_MSOA = os.path.join(local_data_path, "pop_MSOA_orig.csv")
    population_2011 = data_loader.load_MOSA_pop(path_to_csv_MSOA)

    # Get all LADS
    pop_lads_2015_2015 = data_loader.read_scenario_data(path_to_csv_LAD_pop)

    all_msoas = []
    for lad, msoas in population_2011.items():
        all_msoas += list(msoas.keys())

    gva_dummy_msoa_lad = defaultdict(dict)

    # all lads
    all_lads = list(pop_lads_2015_2015[2015].keys())

    # ----------------------------------------------------
    # Iterate MSOA and add population
    # ----------------------------------------------------
    for msoa in all_msoas:

        for lad, msoas in population_2011.items():
            msoas_of_lad = list(msoas.keys())
            if msoa in msoas_of_lad:
                lad_match = lad
                break

        # Remap
        lad_match_remap = remap_lads(lad_match)

        # Remove lad
        for lad_position, lad in enumerate(all_lads):
            if lad == lad_match_remap:
                all_lads.pop(lad_position)
                break


        for year in pop_lads_2015_2015.keys():
            gva_dummy_msoa_lad[year][msoa] = dummy_gva

    # ----------------------------------------------------
    # Iterate all lads and add population for northern ireland and scotland
    # ----------------------------------------------------
    for lad in all_lads:
        for year in pop_lads_2015_2015.keys():
            gva_dummy_msoa_lad[year][lad] = dummy_gva

    # Write msoa to file
    rows = [("region", "year", "value", "interval")]

    for year, msoas in gva_dummy_msoa_lad.items():
        for msoa, gva in msoas.items():
            rows.append([msoa, year, gva, 1])

    write_data.create_csv_file(path_out, rows)
    print("Finished gva script")



"""Read in model results and plot results
"""
import os

from energy_demand.read_write import data_loader, read_data
from energy_demand.basic import date_prop
from energy_demand.plotting import plotting_results, result_mapping
from energy_demand.basic import basic_functions
from energy_demand.basic import lookup_tables

def main(
        path_data_ed,
        path_shapefile_input,
        plot_crit_dict,
        base_yr,
        comparison_year
    ):
    """Read in all results and plot PDFs

    Arguments
    ----------
    path_data_ed : str
        Path to results
    path_shapefile_input : str
        Path to shapefile
    plot_crit_dict : dict
        Criteria to select plots to plot
    base_yr : int
        Base year
    comparison_year : int
        Year to generate comparison plots
    """
    print("...Start creating plots")
    data = {}

    # ---------------------------------------------------------
    # Iterate folders and read out all weather years and stations
    # ---------------------------------------------------------
    to_ignores = [
        'model_run_pop',
        'PDF_validation']

    endings_to_ignore = [
        '.pdf',
        '.txt',
        '.ini']

    all_result_folders = os.listdir(path_data_ed)
    paths_folders_result = []

    for result_folder in all_result_folders:
        if result_folder not in to_ignores and result_folder[-4:] not in endings_to_ignore:
            paths_folders_result.append(
                os.path.join(path_data_ed, result_folder))

    ####################################################################
    # Calculate results for every weather year
    ####################################################################
    for path_result_folder in paths_folders_result:
        
        print("-----------------------")
        print("path_result_folder: " + str(path_result_folder))
        print("-----------------------")
        # Simulation information is read in from .ini file for results
        data['enduses'], data['assumptions'], data['regions'] = data_loader.load_ini_param(
            os.path.join(path_data_ed))

        # ------------------
        # Load necessary inputs for read in
        # ------------------
        data = {}
        data['local_paths'] = data_loader.get_local_paths(
            path_result_folder)
        data['result_paths'] = basic_functions.get_result_paths(
            os.path.join(path_result_folder))
        data['lookups'] = lookup_tables.basic_lookups()

        # ---------------
        # Folder cleaning
        # ---------------
        basic_functions.del_previous_setup(data['result_paths']['data_results_PDF'])
        basic_functions.del_previous_setup(data['result_paths']['data_results_shapefiles'])
        basic_functions.create_folder(data['result_paths']['data_results_PDF'])
        basic_functions.create_folder(data['result_paths']['data_results_shapefiles'])
        basic_functions.create_folder(data['result_paths']['individual_enduse_lp'])
    
        # Simulation information is read in from .ini file for results
        data['enduses'], data['assumptions'], data['regions'] = data_loader.load_ini_param(
            os.path.join(path_data_ed))

        # Other information is read in
        data['assumptions']['seasons'] = date_prop.get_season(year_to_model=2015)
        data['assumptions']['model_yeardays_daytype'], data['assumptions']['yeardays_month'], data['assumptions']['yeardays_month_days'] = date_prop.get_yeardays_daytype(year_to_model=2015)

        data['scenario_data'] = {}

        data['scenario_data']['population'] = read_data.read_scenaric_population_data(
            os.path.join(path_data_ed, 'model_run_pop'))

        # --------------------------------------------
        # Reading in results from different model runs
        # --------------------------------------------
        results_container = read_data.read_in_results(
            data['result_paths']['data_results_model_run_results_txt'],
            data['assumptions']['seasons'],
            data['assumptions']['model_yeardays_daytype'])

        # ------------------------------
        # Plotting other results
        # ------------------------------
        plotting_results.run_all_plot_functions(
            results_container,
            data['assumptions']['reg_nrs'],
            data['regions'],
            data['lookups'],
            data['result_paths'],
            data['assumptions'],
            data['enduses'],
            plot_crit=plot_crit_dict,
            base_yr=base_yr,
            comparison_year=comparison_year)

        # ------------------------------
        # Plotting spatial results
        # ------------------------------
        if plot_crit_dict['spatial_results']:
            result_mapping.spatial_maps(
                data,
                results_container,
                data['result_paths']['data_results_shapefiles'],
                data['regions'],
                data['lookups']['fueltypes_nr'],
                data['lookups']['fueltypes'],
                path_shapefile_input,
                plot_crit_dict,
                base_yr=base_yr)

    print("===================================")
    print("... finished reading and plotting results")
    print("===================================")



"""Script to add interval on GVA and population file

Run script on 'data' folder in scenarios_not_extracted folder
"""
import os
import pandas as pd
import numpy as np
from energy_demand.basic import lookup_tables
from energy_demand.basic import basic_functions

def run(
        path_to_folder,
        path_MSOA_baseline,
        MSOA_calculations=False,
        geography_name='region',
        scenarios_to_generate=[]
        ):
    """
    path_to_folder : str
        Path to data folder
    path_MSOA_baseline : str
        Path to MSOA file with correct geography in csv
    """
    sectors_to_generate = [2, 3, 4, 5, 6, 8, 9, 29, 11, 12, 10, 15, 14, 19, 17, 40, 41, 28, 35, 23, 27]

    # Get all folders with scenario run results
    all_csv_folders = basic_functions.get_all_folders_files(path_to_folder)

    # Lookup of economic sectors
    LAD_MSOA_lu = lookup_tables.lad_msoa_mapping()

    base_yr = 2015
    end_yr = 2050

    # ---------------------------------------------------------------------------------------------------
    # Create scenario with CONSTANT (2015) population and CONSTANT GVA
    # ---------------------------------------------------------------------------------------------------
    '''
    empty_folder_name = os.path.join(path_to_folder, "constant_pop_gva")
    basic_functions.delete_folder(empty_folder_name)
    os.makedirs(empty_folder_name)
    wrote_out_pop, wroute_out_GVA = False, False #Do not change

    # Creat empty dataframe
    columns = ['timestep', 'sector', 'lad_uk_2016', 'value']
    
    # Get folder with standard scenario to get data for constant scenario
    for folder_name in ['pop-baseline16_econ-c16_fuel-c16']:

        all_files = os.listdir(os.path.join(path_to_folder, folder_name))

        # Scale for every year according to this distribution
        for file_name in all_files:
            filename_split = file_name.split("__")
            
            if (filename_split[0] == "gva_per_head" and filename_split[1] == 'lad_sector.csv') or (
                filename_split[0] == "population" and filename_split[1] == 'lad.csv'):

                    file_path = os.path.join(path_to_folder, folder_name, file_name)
                    print("Change file: " + str(file_path))

                    # Read csv file
                    gp_file = pd.read_csv(file_path)

                    # Replace future pop with 2015 pop
                    gp_file_selection_2015 = gp_file.loc[gp_file['year'] == 2015] #Data of 2015

                    list_with_all_vals = []
                    for year in range(base_yr, end_yr + 1):
                        gp_file_selection_yr = gp_file_selection_2015
                        gp_file_selection_yr['year'] = year
                        list_with_all_vals += gp_file_selection_yr.values.tolist()
    
                    # Save as file
                    new_dataframe = pd.DataFrame(list_with_all_vals, columns=gp_file.columns)
                    file_path_out = os.path.join(empty_folder_name, file_name)
                    new_dataframe.to_csv(file_path_out, index=False) #Index prevents writing index rows

                    # ---
                    # MSOA pop calculation
                    # ----
                    if MSOA_calculations:
                        if (filename_split[0] == "population" and filename_split[1] == 'lad.csv'):
                            
                            # Calculate relative pop percentage of ONS scenarios
                            msoa_principalDF = pd.read_csv(path_MSOA_baseline)

                            msoa_principalDF_selection_2015 = msoa_principalDF.loc[msoa_principalDF['year'] == 2015]
                            
                            # LADs and calculate factor per MSOA
                            factor_msoas = {}
                            for lad, msoas in LAD_MSOA_lu.items():
                                tot_pop_lad = 0
                                for msoa in msoas:
                                    tot_pop_lad += float(msoa_principalDF_selection_2015.loc[msoa_principalDF_selection_2015['region'] == msoa]['value'])
                                for msoa in msoas:
                                    pop_msoa = float(msoa_principalDF_selection_2015.loc[msoa_principalDF_selection_2015['region'] == msoa]['value'])
                                    factor_msoas[msoa] = pop_msoa / tot_pop_lad #calculate fator

                            list_with_all_vals = []

                            # READ csv file
                            gp_file = pd.read_csv(file_path)

                            pop_LADs_2015 = gp_file.loc[gp_file['year'] == 2015]

                            for index, row_lad in gp_file.iterrows():
                                lad = row_lad['region']
                                try:
                                    corresponding_msoas = LAD_MSOA_lu[lad]
                                except KeyError:
                                    # No match for northern ireland
                                    corresponding_msoas = [lad]

                                # Calculate population according to ONS 2015 #pop_LAD = row_lad['value']
                                pop_LAD_2015 = float(pop_LADs_2015.loc[gp_file['region'] == lad]['value']) #Base year pop
                                
                                for msoa_name in corresponding_msoas:
                                    try:
                                        pop_ONS_scale_factor = factor_msoas[msoa_name]
                                    except:
                                        pop_ONS_scale_factor = 1 # If not mapped
                                    
                                    pop_MSOA_ONS_scaled = pop_LAD_2015 * pop_ONS_scale_factor

                                    new_row = {
                                        'region': msoa_name,
                                        "year": row_lad['year'],
                                        "value": pop_MSOA_ONS_scaled,
                                        "interval": row_lad['interval']}

                                    list_with_all_vals.append(new_row)

                            msoaDF = pd.DataFrame(list_with_all_vals, columns=gp_file.columns)
                            file_path_MSOA_out = os.path.join(empty_folder_name, "{}_{}.csv".format(file_name[:-4], "MSOA"))
                            msoaDF.to_csv(file_path_MSOA_out, index=False)

                    wrote_out_pop = True

            elif (filename_split[0] == "gva_per_head" and filename_split[1] == 'lad.csv'):

                    file_path = os.path.join(path_to_folder, folder_name, file_name)
                    print("Change file: " + str(file_path))

                    gp_file = pd.read_csv(file_path)
        
                    # Add new column
                    gp_file['value'] = 1000

                    # Replace future pop with 2015 pop
                    gp_file_selection_2015 = gp_file.loc[gp_file['year'] == 2015] #Data of 2015

                    list_with_all_vals = []
                    for year in range(base_yr, end_yr + 1):
                        gp_file_selection_yr = gp_file_selection_2015
                        gp_file_selection_yr['year'] = year
                        list_with_all_vals += gp_file_selection_yr.values.tolist()

                    new_dataframe = pd.DataFrame(list_with_all_vals, columns=gp_file.columns)

                    # Save as file
                    file_path_out = os.path.join(empty_folder_name, file_name)
                    new_dataframe.to_csv(file_path_out, index=False) #Index prevents writing index rows

                    # -----------------------------------------
                    # MSOA GVA calculations
                    # -----------------------------------------
                    if MSOA_calculations:
                        lads = list(gp_file.loc[gp_file['year'] == 2015]['region'])
                        list_with_all_vals = []

                        for lad in lads:
                            try:
                                corresponding_msoas = LAD_MSOA_lu[lad]
                            except KeyError:
                                corresponding_msoas = lad # No match for northern ireland

                            rows_msoa = gp_file.loc[gp_file['region'] == lad].values

                            for row_msoa in rows_msoa:
                                for msoa_name in corresponding_msoas:
                                    #row_msoa[0] = msoa_name
                                    new_row = {
                                        "region": msoa_name,
                                        "year": row_msoa[1],
                                        "value": row_msoa[2],
                                        "interval": row_msoa[3]}
                                    list_with_all_vals.append(new_row)
                                    #msoaDF = msoaDF.append(new_row, ignore_index=True)
                        # Convert list to dataframe
                        msoaDF = pd.DataFrame(list_with_all_vals, columns=gp_file.columns)
                        file_path_MSOA_out = os.path.join(empty_folder_name, "{}_{}.csv".format(file_name[:-4], "MSOA"))
                        msoaDF.to_csv(file_path_MSOA_out, index=False)

                    wroute_out_GVA = True
            else:
                pass
        if wrote_out_pop == True and wroute_out_GVA == True:
            break

    print("... finished generating CONSTANT scenario")'''
    # ---------------------------------------------------------------------------------------------------
    # Add interval and create individual GVA data for selected sectors
    # ---------------------------------------------------------------------------------------------------
    columns = ['timestep', 'sectors', geography_name]

    # Get all folders with scenario run results (name of folder is scenario)
    all_csv_folders_walk = os.walk(path_to_folder)
    for root, dirnames, filenames in all_csv_folders_walk:
        all_csv_folders = dirnames
        break

    for folder_name in all_csv_folders:
        all_files = os.listdir(os.path.join(path_to_folder, folder_name))
        if (scenarios_to_generate == []) or (folder_name in scenarios_to_generate):
            print("folder name: " + str(folder_name), flush=True)
            for file_name in all_files:
                filename_split = file_name.split("__")
                var_name = filename_split[0]
                if (var_name == "gva_per_head" and filename_split[1] == 'lad_sector.csv') or (
                    var_name == "population" and filename_split[1] == 'lad.csv') or (
                        var_name == "gva_per_head" and filename_split[1] == 'lad.csv'):
                    
                    try:

                        file_path = os.path.join(path_to_folder, folder_name, file_name)
                        print("file_path " + str(file_path))
                        gp_file = pd.read_csv(file_path)

                        gp_file['year'] = gp_file['year'].astype(int)

                        # Drop all rows with alls NaN entries
                        gp_file = gp_file[np.isfinite(gp_file['value'])]

                        # Select all entries with matching years
                        gp_file = gp_file.loc[gp_file['year'].isin(range(base_yr, end_yr + 1))]

                        # Rename columns
                        gp_file = gp_file.rename(index=str, columns={"year": "timestep"})
                        gp_file = gp_file.rename(index=str, columns={"region": geography_name})
                        gp_file = gp_file.rename(index=str, columns={"value": var_name})

                        gp_file.to_csv(file_path, index=False)

                        # ---  
                        # MSOA pop calculation
                        # ----
                        if MSOA_calculations:
                            if (filename_split[0] == "population" and filename_split[1] == 'lad.csv'):
                                
                                # Calculate relative pop percentage of ONS scenarios
                                msoa_principalDF = pd.read_csv(path_MSOA_baseline)
                                msoa_principalDF_selection_2015 = msoa_principalDF.loc[msoa_principalDF['year'] == 2015]
                                
                                # LADs and calculate factor per MSOA
                                factor_msoas = {}
                                for lad, msoas in LAD_MSOA_lu.items():
                                    tot_pop_lad = 0
                                    for msoa in msoas:
                                        tot_pop_lad += float(msoa_principalDF_selection_2015.loc[msoa_principalDF_selection_2015['region'] == msoa]['value'])
                                    for msoa in msoas:
                                        pop_msoa = float(msoa_principalDF_selection_2015.loc[msoa_principalDF_selection_2015['region'] == msoa]['value'])
                                        factor_msoas[msoa] = pop_msoa / tot_pop_lad #calculate fator

                                list_with_all_vals = []
                                gp_file = pd.read_csv(file_path)

                                for index, row_lad in gp_file.iterrows():
                                    lad = row_lad['region']
                                    try:
                                        corresponding_msoas = LAD_MSOA_lu[lad]
                                    except KeyError:
                                        corresponding_msoas = [lad] # No match for northern ireland

                                    # Calculate population according to ONS 2015
                                    pop_LAD = row_lad['value']

                                    for msoa_name in corresponding_msoas:
                                        try:
                                            pop_ONS_scale_factor = factor_msoas[msoa_name]
                                        except:
                                            pop_ONS_scale_factor = 1 # If not mapped
                                        
                                        pop_MSOA_ONS_scaled = pop_LAD * pop_ONS_scale_factor

                                        new_row = {
                                            'region': msoa_name,
                                            "timestep": row_lad['year'],
                                            "value": pop_MSOA_ONS_scaled}

                                        list_with_all_vals.append(new_row)

                                msoaDF = pd.DataFrame(list_with_all_vals, columns=gp_file.columns)
                                file_path_MSOA_out = os.path.join(path_to_folder, folder_name, "{}_{}.csv".format(file_name[:-4], "MSOA"))
                                msoaDF.to_csv(file_path_MSOA_out, index=False)
                    except:
                        print("... error in preparing data")
                        pass
                else:
                    pass

                # ----------------------------------------------------------
                # Script to generate sectors file
                # ----------------------------------------------------------
                if (filename_split[0] == "gva_per_head" and filename_split[1] == 'lad_sector.csv'):
                    
                    #try:
                        file_path = os.path.join(path_to_folder, folder_name, file_name)
                        df = pd.read_csv(file_path)

                        df = df.rename(index=str, columns={"year": "timestep"})
                        df = df.rename(index=str, columns={"economic_sector__gor": "sectors"})
                        df = df.rename(index=str, columns={"gva_per_head": "gva_per_sector"})

                        # Drop columns
                        try:
                            df = df.drop('interval', 1)
                        except:
                            pass
                        # Select all entries with matching sectors
                        df = df.loc[df['sectors'].isin(sectors_to_generate)]

                        # Select all entries with matching years
                        df = df.loc[df['timestep'].isin(range(base_yr, end_yr + 1))]

                        # Write to csv
                        file_path_sectors = os.path.join(path_to_folder, folder_name, "gva_per_head__lad_sectors.csv")
                        df.to_csv(file_path_sectors, index=False)
                    #except:
                    #    pass #Error

                # -----------------------------------------
                # MSOA GVA calculations
                # -----------------------------------------
                if MSOA_calculations:
                    if (filename_split[0] == "gva_per_head" and filename_split[1] == 'lad.csv'):
                        
                        try:
                            list_with_all_vals = []

                            file_path = os.path.join(path_to_folder, folder_name, file_name)
                            lads = list(gp_file.loc[gp_file['year'] == 2015]['region'])

                            for lad in lads:
                                try:
                                    corresponding_msoas = LAD_MSOA_lu[lad]
                                except KeyError:
                                    # No match for northern ireland
                                    corresponding_msoas = [lad]

                                rows_msoa = gp_file.loc[gp_file['region'] == lad]

                                for index, row_msoa in rows_msoa.iterrows():
                                    for msoa_name in corresponding_msoas:
                                        new_row = {
                                            "region": msoa_name,
                                            "timestep": row_msoa['year'],
                                            "value": row_msoa['value']}
                                        list_with_all_vals.append(new_row)

                            # Convert list to dataframe
                            msoaDF = pd.DataFrame(list_with_all_vals, columns=gp_file.columns)

                            msoaDF = msoaDF.rename(index=str, columns={"region": geography_name})

                            file_path_MSOA_out = os.path.join(path_to_folder, folder_name, "{}_{}.csv".format(file_name[:-4], "MSOA"))
                            msoaDF.to_csv(file_path_MSOA_out, index=False)
                        except:
                            pass

    print("----------")
    print("Finished preparing NISMOD GVA and population files")
    print("----------")



"""testing
"""
import numpy as np
from energy_demand.basic import basic_functions

def test_get_month_from_string():
    """Testing
    """
    assert basic_functions.get_month_from_string("Jan") == 1
    assert basic_functions.get_month_from_string("Feb") == 2
    assert basic_functions.get_month_from_string("Mar") == 3
    assert basic_functions.get_month_from_string("Apr") == 4
    assert basic_functions.get_month_from_string("May") == 5
    assert basic_functions.get_month_from_string("Jun") == 6
    assert basic_functions.get_month_from_string("Jul") == 7
    assert basic_functions.get_month_from_string("Aug") == 8
    assert basic_functions.get_month_from_string("Sep") == 9
    assert basic_functions.get_month_from_string("Oct") == 10
    assert basic_functions.get_month_from_string("Nov") == 11
    assert basic_functions.get_month_from_string("Dec") == 12

def test_array_to_dict():
    """Testing
    """
    result_array = np.ones((3)) + 2
    result_array[2] = 44

    regions = ["A", "B", "C"]
    result = basic_functions.array_to_dict(result_array, regions)

    expected = {"A": 3, "B": 3, "C": 44}
    assert result == expected
